"""
Utilities for working with rating matrices.
"""

from collections import namedtuple
import logging

import pandas as pd
import numpy as np
import scipy.sparse as sps

_logger = logging.getLogger(__name__)

RatingMatrix = namedtuple('RatingMatrix', ['matrix', 'users', 'items'])
RatingMatrix.__doc__ = """
A rating matrix with associated indices.

Attributes:
    matrix(CSR or scipy.sparse.csr_matrix):
        The rating matrix, with users on rows and items on columns.
    users(pandas.Index): mapping from user IDs to row numbers.
    items(pandas.Index): mapping from item IDs to column numbers.
"""

__impl_mod = None


def _impl_mod():
    global __impl_mod
    if __impl_mod is None:
        from . import native
        __impl_mod = native
    return __impl_mod


def mkl_ops():
    """
    Import and return the MKL operations module.  This is only for internal use.
    """
    try:
        from . import _mkl_ops
        if _mkl_ops.clib:
            return _mkl_ops
        else:
            _logger.debug('MKL imported but not available')
            return None
    except ImportError as e:
        _logger.debug('MKL not importable: %s', e)
        return None


# This class lives here, un-JITted, so we avoid Numba compiles until necessary
class _CSR:
    """
    Internal implementation class for :py:class:`CSR`. If you work with CSRs from Numba,
    you will use a :func:`numba.jitclass`-ed version of this.

    Note that the ``values`` array is always present (unlike the Python shim), but is
    zero-length if no values are present.  This eases Numba type-checking.
    """
    def __init__(self, nrows, ncols, nnz, ptrs, inds, vals):
        self.nrows = nrows
        self.ncols = ncols
        self.nnz = nnz
        self.rowptrs = ptrs
        self.colinds = inds
        if vals is not None:
            self.values = vals
        else:
            self.values = np.zeros(0)

    def row(self, row):
        sp = self.rowptrs[row]
        ep = self.rowptrs[row + 1]

        v = np.zeros(self.ncols)
        cols = self.colinds[sp:ep]
        if self.values.size == 0:
            v[cols] = 1
        else:
            v[cols] = self.values[sp:ep]

        return v

    def row_extent(self, row):
        sp = self.rowptrs[row]
        ep = self.rowptrs[row+1]
        return (sp, ep)

    def row_cs(self, row):
        sp = self.rowptrs[row]
        ep = self.rowptrs[row + 1]

        return self.colinds[sp:ep]

    def row_vs(self, row):
        sp = self.rowptrs[row]
        ep = self.rowptrs[row + 1]

        if self.values.size == 0:
            return np.full(ep - sp, 1.0)
        else:
            return self.values[sp:ep]

    def rowinds(self):
        ris = np.zeros(self.nnz, np.intc)
        for i in range(self.nrows):
            sp, ep = self.row_extent(i)
            ris[sp:ep] = i
        return ris


def _csr_delegate(name):
    def func(self):
        return getattr(self._N, name)

    return property(func)


class CSR:
    """
    Simple compressed sparse row matrix.  This is like :py:class:`scipy.sparse.csr_matrix`, with
    a couple of useful differences:

    * It is backed by a Numba jitclass, so it can be directly used from Numba-optimized functions.
    * The value array is optional, for cases in which only the matrix structure is required.
    * The value array, if present, is always double-precision.

    You generally don't want to create this class yourself with the constructor.  Instead, use one
    of its class methods.

    If you need to pass an instance off to a Numba-compiled function, use :py:attr:`N`::

        _some_numba_fun(csr.N)

    We use the indirection between this and the Numba jitclass so that the main CSR implementation
    can be pickled, and so that we can have class and instance methods that are not compatible with
    jitclass but which are useful from interpreted code.

    Attributes:
        N(_CSR): the Numba jitclass backing (has the same attributes and most methods).
        nrows(int): the number of rows.
        ncols(int): the number of columns.
        nnz(int): the number of entries.
        rowptrs(numpy.ndarray): the row pointers.
        colinds(numpy.ndarray): the column indices.
        values(numpy.ndarray): the values
    """
    __slots__ = ['_N']

    def __init__(self, nrows=None, ncols=None, nnz=None, ptrs=None, inds=None, vals=None, N=None):
        if N is not None:
            self._N = N
        else:
            self._N = _CSR(nrows, ncols, nnz, ptrs, inds, vals)

    @classmethod
    def empty(cls, shape, row_nnzs, *, rpdtype=np.intc):
        """
        Create an empty CSR matrix.

        Args:
            shape(tuple): the array shape (rows,cols)
            row_nnzs(array-like): the number of nonzero entries for each row
        """
        nrows, ncols = shape
        assert len(row_nnzs) == nrows

        nnz = np.sum(row_nnzs, dtype=np.int64)
        rowptrs = np.zeros(nrows + 1, dtype=rpdtype)
        rowptrs[1:] = np.cumsum(row_nnzs, dtype=rpdtype)
        colinds = np.full(nnz, -1, dtype=np.intc)
        values = np.full(nnz, np.nan)

        return cls(nrows, ncols, nnz, rowptrs, colinds, values)

    @classmethod
    def from_coo(cls, rows, cols, vals, shape=None, rpdtype=np.intc):
        """
        Create a CSR matrix from data in COO format.

        Args:
            rows(array-like): the row indices.
            cols(array-like): the column indices.
            vals(array-like): the data values; can be ``None``.
            shape(tuple): the array shape, or ``None`` to infer from row & column indices.
        """
        if shape is not None:
            nrows, ncols = shape
            assert np.max(rows) < nrows
            assert np.max(cols) < ncols
        else:
            nrows = np.max(rows) + 1
            ncols = np.max(cols) + 1

        nnz = len(rows)
        assert len(cols) == nnz
        assert vals is None or len(vals) == nnz

        rowptrs = np.zeros(nrows + 1, dtype=rpdtype)
        align = np.full(nnz, -1, dtype=rpdtype)

        _impl_mod()._csr_align(rows, nrows, rowptrs, align)

        cols = cols[align].copy()
        vals = vals[align].copy() if vals is not None else None

        return cls(nrows, ncols, nnz, rowptrs, cols, vals)

    @classmethod
    def from_scipy(cls, mat, copy=True):
        """
        Convert a scipy sparse matrix to an internal CSR.

        Args:
            mat(scipy.sparse.spmatrix): a SciPy sparse matrix.
            copy(bool): if ``False``, reuse the SciPy storage if possible.

        Returns:
            CSR: a CSR matrix.
        """
        if not sps.isspmatrix_csr(mat):
            mat = mat.tocsr(copy=copy)
        rp = np.require(mat.indptr, np.intc, 'C')
        if copy and rp is mat.indptr:
            rp = rp.copy()
        cs = np.require(mat.indices, np.intc, 'C')
        if copy and cs is mat.indices:
            cs = cs.copy()
        vs = mat.data.copy() if copy else mat.data
        return cls(mat.shape[0], mat.shape[1], mat.nnz, rp, cs, vs)

    def to_scipy(self):
        """
        Convert a CSR matrix to a SciPy :py:class:`scipy.sparse.csr_matrix`.  Avoids copying
        if possible.

        Args:
            self(CSR): A CSR matrix.

        Returns:
            scipy.sparse.csr_matrix:
                A SciPy sparse matrix with the same data.
        """
        values = self.values
        if values is None:
            values = np.full(self.nnz, 1.0)
        return sps.csr_matrix((values, self.colinds, self.rowptrs), shape=(self.nrows, self.ncols))

    @property
    def N(self):
        """
        Get the native backing array.
        """
        # we do this *lazily*.
        n = self._N
        if isinstance(n, _CSR):
            # This is not yet upgraded
            impl = _impl_mod()
            if n.rowptrs.dtype == np.int64:
                self._N = impl._CSR64(n.nrows, n.ncols, n.nnz, n.rowptrs, n.colinds, n.values)
            else:
                self._N = impl._CSR(n.nrows, n.ncols, n.nnz, n.rowptrs, n.colinds, n.values)
            # make sure types behave how we expect
            assert impl.n.config.DISABLE_JIT or not isinstance(self._N, _CSR)
        return self._N

    nrows = _csr_delegate('nrows')
    ncols = _csr_delegate('ncols')
    nnz = _csr_delegate('nnz')
    rowptrs = _csr_delegate('rowptrs')
    colinds = _csr_delegate('colinds')

    @property
    def values(self):
        if self._N.values is not None and self._N.values.size:
            return self._N.values
        else:
            return None

    @values.setter
    def values(self, vs: np.ndarray):
        if vs is not None:
            if not isinstance(vs, np.ndarray):
                raise TypeError('values not an ndarray')
            if vs.ndim != 1:
                raise ValueError('values has {} dimensions, expected 1'.format(vs.ndims))
            if vs.shape[0] < self.nnz:
                s = 'values has only {} entries (expected at least {})'
                raise ValueError(s.format(vs.shape[0], self.nnz))

            vs = vs[:self.nnz]
            vs = np.require(vs, 'f8')
            self._N.values = vs
        else:
            self._N.values = np.zeros(0)

    def subset_rows(self, begin, end):
        """
        Subset the rows in this matrix.
        """
        impl = _impl_mod()
        return CSR(N=impl._subset_rows(self.N, begin, end))

    def rowinds(self) -> np.ndarray:
        """
        Get the row indices from this array.  Combined with :py:attr:`colinds` and
        :py:attr:`values`, this can form a COO-format sparse matrix.
        """
        # we have to upgrade to Numba for this one - too slow in Python
        return self.N.rowinds()

    def row(self, row):
        """
        Return a row of this matrix as a dense ndarray.

        Args:
            row(int): the row index.

        Returns:
            numpy.ndarray: the row, with 0s in the place of missing values.
        """
        return self._N.row(row)

    def row_extent(self, row):
        """
        Get the extent of a row in the underlying column index and value arrays.

        Args:
            row(int): the row index.

        Returns:
            tuple: ``(s, e)``, where the row occupies positions :math:`[s, e)` in the
            CSR data.
        """
        return self._N.row_extent(row)

    def row_cs(self, row):
        """
        Get the column indcies for the stored values of a row.
        """
        return self._N.row_cs(row)

    def row_vs(self, row):
        """
        Get the stored values of a row.
        """
        return self._N.row_vs(row)

    def row_nnzs(self):
        """
        Get a vector of the number of nonzero entries in each row.

        .. note:: This method is not available from Numba.

        Returns:
            numpy.ndarray: the number of nonzero entries in each row.
        """
        return np.diff(self.rowptrs)

    def normalize_rows(self, normalization):
        """
        Normalize the rows of the matrix.

        .. note:: The normalization *ignores* missing values instead of treating
                  them as 0.

        .. note:: This method is not available from Numba.

        Args:
            normalization(str):
                The normalization to perform. Can be one of:

                * ``'center'`` - center rows about the mean
                * ``'unit'`` - convert rows to a unit vector

        Returns:
            numpy.ndarray:
                The normalization values for each row.
        """
        impl = _impl_mod()
        if normalization == 'center':
            return impl._center_rows(self.N)
        elif normalization == 'unit':
            return impl._unit_rows(self.N)
        else:
            raise ValueError('unknown normalization: ' + normalization)

    def transpose(self, values=True):
        """
        Transpose a CSR matrix.

        .. note:: This method is not available from Numba.

        Args:
            values(bool): whether to include the values in the transpose.

        Returns:
            CSR: the transpose of this matrix (or, equivalently, this matrix in CSC format).
        """
        impl = _impl_mod()
        n_rows = self.rowinds()
        rows = self.colinds
        n_vs = self.values if values else None
        if n_vs is not None:
            n_vs = n_vs.copy()

        rowptrs = impl._csr_align_inplace((self.ncols, self.nrows), rows, n_rows, n_vs)
        if self.rowptrs.dtype == np.int32:
            rowptrs = rowptrs.astype(np.int32)

        return CSR(self.ncols, self.nrows, self.nnz, rowptrs, n_rows, n_vs)

    def filter_nnzs(self, filt):
        """
        Filter the values along the full NNZ axis.

        Args:
            filt(ndarray):
                a logical array of length :attr:`nnz` that indicates the values to keep.

        Returns:
            CSR: The filtered sparse matrix.
        """
        if len(filt) != self.nnz:
            raise ValueError('filter has length %d, expected %d' % (len(filt), self.nnz))
        rps2 = np.zeros_like(self.rowptrs)
        for i in range(self.nrows):
            sp, ep = self.row_extent(i)
            rlen = np.sum(filt[sp:ep])
            rps2[i+1] = rps2[i] + rlen

        nnz2 = rps2[-1]
        assert nnz2 == np.sum(filt)

        cis2 = self.colinds[filt]
        vs = self.values
        vs2 = None if vs is None else vs[filt]

        return CSR(self.nrows, self.ncols, nnz2, rps2, cis2, vs2)

    def __str__(self):
        return '<CSR {}x{} ({} nnz)>'.format(self.nrows, self.ncols, self.nnz)

    def __repr__(self):
        repr = '<CSR {}x{} ({} nnz)'.format(self.nrows, self.ncols, self.nnz)
        repr += ' {\n'
        repr += '  rowptrs={}\n'.format(self.rowptrs)
        repr += '  colinds={}\n'.format(self.colinds)
        repr += '  values={}\n'.format(self.values)
        repr += '}'
        return repr

    def __getstate__(self):
        return dict(shape=(self.nrows, self.ncols), nnz=self.nnz,
                    rowptrs=self.rowptrs, colinds=self.colinds, values=self.values)

    def __setstate__(self, state):
        nrows, ncols = state['shape']
        nnz = state['nnz']
        rps = state['rowptrs']
        cis = state['colinds']
        vs = state['values']
        self._N = _CSR(nrows, ncols, nnz, rps, cis, vs)


def sparse_ratings(ratings, scipy=False, *, users=None, items=None):
    """
    Convert a rating table to a sparse matrix of ratings.

    Args:
        ratings(pandas.DataFrame): a data table of (user, item, rating) triples.
        scipy: if ``True``, return a SciPy matrix instead of :py:class:`CSR`.
        users(pandas.Index): an index of user IDs.
        items(pandas.Index): an index of items IDs.

    Returns:
        RatingMatrix:
            a named tuple containing the sparse matrix, user index, and item index.
    """
    if users is None:
        users = pd.Index(np.unique(ratings.user), name='user')

    if items is None:
        items = pd.Index(np.unique(ratings.item), name='item')

    _logger.debug('creating matrix with %d ratings for %d items by %d users',
                  len(ratings), len(items), len(users))

    row_ind = users.get_indexer(ratings.user).astype(np.intc)
    if np.any(row_ind < 0):
        raise ValueError('provided user index does not cover all users')
    col_ind = items.get_indexer(ratings.item).astype(np.intc)
    if np.any(col_ind < 0):
        raise ValueError('provided item index does not cover all users')

    if 'rating' in ratings.columns:
        vals = np.require(ratings.rating.values, np.float64)
    else:
        vals = None

    matrix = CSR.from_coo(row_ind, col_ind, vals, (len(users), len(items)))

    if scipy:
        matrix = matrix.to_scipy()

    return RatingMatrix(matrix, users, items)



"""
Test utilities for LKPY tests.
"""

import os
import os.path
import logging
from contextlib import contextmanager

import numpy as np
from .. import matrix

import pytest
from hypothesis import given, assume
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph

from lenskit.datasets import MovieLens, ML100K

_log = logging.getLogger(__name__)

ml_test = MovieLens('data/ml-latest-small')
ml100k = ML100K('data/ml-100k')


@st.composite
def csrs(draw, nrows=None, ncols=None, nnz=None, values=None):
    if ncols is None:
        ncols = draw(st.integers(5, 100))
    elif not isinstance(ncols, int):
        ncols = draw(ncols)

    if nrows is None:
        nrows = draw(st.integers(5, 100))
    elif not isinstance(nrows, int):
        nrows = draw(nrows)

    if nnz is None:
        nnz = draw(st.integers(10, nrows * ncols // 2))
    elif not isinstance(nnz, int):
        nnz = draw(nnz)

    coords = draw(nph.arrays(np.int32, nnz, elements=st.integers(0, nrows*ncols - 1), unique=True))
    rows = np.mod(coords, nrows, dtype=np.int32)
    cols = np.floor_divide(coords, nrows, dtype=np.int32)
    if values is None:
        values = draw(st.booleans())
    if values:
        rng = draw(st.randoms())
        vals = np.array([rng.normalvariate(0, 1) for i in range(nnz)])
    else:
        vals = None
    return matrix.CSR.from_coo(rows, cols, vals, (nrows, ncols))


@contextmanager
def set_env_var(var, val):
    "Set an environment variable & restore it."
    is_set = var in os.environ
    if is_set:
        old_val = os.environ[var]
    try:
        if val is None:
            if is_set:
                del os.environ[var]
        else:
            os.environ[var] = val
        yield
    finally:
        if is_set:
            os.environ[var] = old_val
        elif val is not None:
            del os.environ[var]


wantjit = pytest.mark.skipif('NUMBA_DISABLE_JIT' in os.environ,
                             reason='JIT required')



import logging
import pickle
from . import sharing_mode, PersistedModel

try:
    import multiprocessing.shared_memory as shm
    SHM_AVAILABLE = True
except ImportError:
    SHM_AVAILABLE = False

if pickle.HIGHEST_PROTOCOL < 5:
    import pickle5 as pickle

_log = logging.getLogger(__name__)


def persist_shm(model, dir=None):
    """
    Persist a model using binpickle.

    Args:
        model: The model to persist.
        dir: The temporary directory for persisting the model object.

    Returns:
        PersistedModel: The persisted object.
    """
    if shm is None:
        raise ImportError('multiprocessing.shared_memory')

    buffers = []

    with sharing_mode():
        data = pickle.dumps(model, protocol=5, buffer_callback=buffers.append)

    total_size = sum(memoryview(b).nbytes for b in buffers)
    _log.info('serialized %s to %d pickle bytes with %d buffers of %d bytes',
              model, len(data), len(buffers), total_size)

    if buffers:
        # blit the buffers to the SHM block
        memory = shm.SharedMemory(create=True, size=total_size)
        cur_offset = 0
        blocks = []
        for buf in buffers:
            ba = buf.raw()
            blen = ba.nbytes
            bend = cur_offset + blen
            _log.debug('saving %d bytes', blen)
            memory.buf[cur_offset:bend] = ba
            blocks.append((cur_offset, bend))
            cur_offset = bend
    else:
        memory = None
        blocks = []

    return SHMPersisted(data, memory, blocks)


class SHMPersisted(PersistedModel):
    buffers = []
    _model = None
    memory = None

    def __init__(self, data, memory, blocks):
        self.pickle_data = data
        self.blocks = blocks
        self.memory = memory
        self.shm_name = memory.name if memory is not None else None
        self.is_owner = True

    def get(self):
        if self._model is None:
            _log.debug('loading model from shared memory')
            shm = self._open()
            buffers = []
            for bs, be in self.blocks:
                buffers.append(shm.buf[bs:be])

            self._model = pickle.loads(self.pickle_data, buffers=buffers)

        return self._model

    def close(self, unlink=True):
        self._model = None

        _log.debug('releasing SHM buffers')
        self.buffers = None
        if self.memory is not None:
            self.memory.close()
            if self.is_owner:
                self.memory.unlink()
                self.is_owner = False
            self.memory = None

    def _open(self):
        if self.shm_name and not self.memory:
            self.memory = shm.SharedMemory(name=self.shm_name)
        return self.memory

    def __getstate__(self):
        return {
            'pickle_data': self.pickle_data,
            'blocks': self.blocks,
            'shm_name': self.shm_name,
            'is_owner': True if self.is_owner == 'transfer' else False
        }

    def __setstate__(self, state):
        self.__dict__.update(state)
        if self.is_owner:
            _log.debug('opening shared buffers after ownership transfer')
            self._open()



import logging
import math

import pandas as pd
import numpy as np
from numba import njit

try:
    import tensorflow as tf
    import tensorflow.keras as k
except ImportError:
    tf = None

from lenskit import util
from lenskit.matrix import sparse_ratings
from .. import Predictor
from .util import init_tf_rng, check_tensorflow

_log = logging.getLogger(__name__)


@njit
def _neg_sample(mat, uv):
    """
    Sample the negative examples.  For each user in uv, it samples an item that
    they have not rated using rejection sampling.

    While this is embarassingly parallel, we do not parallelize because TensorFlow
    will request multiple batches in parallel.
    """
    n = len(uv)
    ni = mat.ncols
    jv = np.empty(n, dtype=np.int32)
    sc = np.ones(n, dtype=np.int32)

    for i in range(n):
        u = uv[i]
        used = mat.row_cs(u)
        j = np.random.randint(0, ni)
        while np.any(used == j):
            j = np.random.randint(0, ni)
            sc[i] = sc[i] + 1
        jv[i] = j

    return jv, sc


if tf is not None:
    class BprLoss(k.losses.Loss):
        def call(self, y_true, y_pred):
            return k.backend.mean(-tf.math.log_sigmoid(y_pred))

    class BprInputs(k.utils.Sequence):
        def __init__(self, urmat, batch_size, neg_count, rng):
            super().__init__()
            self.n_items = urmat.ncols
            self.matrix = urmat
            self.users = urmat.rowinds()
            self.items = urmat.colinds
            self.rng = rng
            self.batch_size = batch_size
            self.neg_count = neg_count
            self.permutation = np.arange(self.matrix.nnz, dtype='i4')
            self.targets = np.ones(batch_size * neg_count)
            rng.shuffle(self.permutation)

        def __len__(self):
            return math.ceil(self.matrix.nnz / self.batch_size)

        def __getitem__(self, idx):
            _log.debug('preparing batch %d', idx)
            start = idx * self.batch_size
            end = min(start + self.batch_size, self.matrix.nnz)
            picked = self.permutation[start:end]
            if self.neg_count > 1:
                # expand picked size to sample more items
                picked = np.concatenate([picked for i in range(self.neg_count)])
            assert len(picked) == self.neg_count * (end - start)
            uv = self.users[picked]
            iv = self.items[picked]
            jv, j_samps = _neg_sample(self.matrix.N, uv)
            assert all(jv < self.n_items)
            _log.debug('max sample count: %d', j_samps.max())
            return [uv.astype(np.int32),
                    iv.astype(np.int32),
                    jv.astype(np.int32)], self.targets[:len(picked)]

        def on_epoch_end(self):
            _log.info('re-shuffling')
            self.rng.shuffle(self.permutation)


class BPR(Predictor):
    """
    Bayesian Personalized Ranking with matrix factorization, optimized with TensorFlow.

    This is a basic TensorFlow implementation of the BPR algorithm _[BPR].

    .. _[BPR]:
        Rendle, S. et al. (2009) ‘BPR: Bayesian Personalized Ranking from Implicit
        Feedback’, in *Proceedings of the Twenty-Fifth Conference on Uncertainty in
        Artificial Intelligence*. AUAI Press (UAI ’09), pp. 452–461.

    User and item embedding matrices are regularized with :math:`L_2` regularization,
    governed by a regularization term :math:`\\lambda`.  Regularizations for the user
    and item embeddings are then computed as follows:

    .. math::
        \\lambda_u = \\lambda / |U| \\\\
        \\lambda_i = \\lambda / |I| \\\\

    This rescaling allows the regularization term to be independent of the number of
    users and items.

    Because the model is relatively simple, optimization works best with large
    batch sizes.

    Args:
        features(int): The number of latent features to learn.
        epochs(int): The number of epochs to train.
        batch_size(int):
            The Keras batch size.  This is the number of **positive** examples
            to sample in each batch.  If ``neg_count`` is greater than 1, the
            batch size will be similarly multipled.
        reg(double):
            The regularization term for the embedding vectors.
        neg_count(int):
            The number of negative examples to sample for each positive one.
        rng_spec:
            The random number generator initialization.

    Attributes:
        model: The Keras model.
    """

    model = None

    def __init__(self, features=50, *, epochs=5, batch_size=10000,
                 reg=0.02, neg_count=1, rng_spec=None):
        check_tensorflow()
        self.features = features
        self.epochs = epochs
        self.batch_size = batch_size
        self.reg = reg
        self.neg_count = neg_count
        self.rng_spec = rng_spec

    def fit(self, ratings, **kwargs):
        timer = util.Stopwatch()
        rng = util.rng(self.rng_spec)

        matrix, users, items = sparse_ratings(ratings[['user', 'item']])

        _log.info('[%s] setting up model', timer)
        train, model = self._build_model(len(users), len(items))

        _log.info('[%s] preparing training dataset', timer)
        train_data = BprInputs(matrix, self.batch_size, self.neg_count, rng)

        _log.info('[%s] training model', timer)
        train.fit(train_data, epochs=self.epochs)

        _log.info('[%s] model finished', timer)

        self.user_index_ = users
        self.item_index_ = items
        self.model = model

        return self

    def _build_model(self, n_users, n_items):
        n_features = self.features
        _log.info('configuring TensorFlow model for %d features from %d users and %d items',
                  n_features, n_users, n_items)

        init_tf_rng(self.rng_spec)

        # User input layer
        u_input = k.Input(shape=(1,), dtype='int32', name='user')
        # User embedding layer.
        u_reg = k.regularizers.l2(self.reg / n_users)
        u_embed = k.layers.Embedding(input_dim=n_users, output_dim=n_features, input_length=1,
                                     embeddings_regularizer=u_reg,
                                     embeddings_initializer='random_normal',
                                     name='user-embed')
        # The embedding layer produces an extra dimension. Remove it.
        u_flat = k.layers.Flatten(name='user-vector')(u_embed(u_input))

        # Do the same thing for items
        i_input = k.Input(shape=(1,), dtype='int32', name='item')
        i_reg = k.regularizers.l2(self.reg / n_items)
        i_embed = k.layers.Embedding(input_dim=n_items, output_dim=n_features, input_length=1,
                                     embeddings_regularizer=i_reg,
                                     embeddings_initializer='random_normal',
                                     name='item-embed')
        i_flat = k.layers.Flatten(name='item-vector')(i_embed(i_input))

        # we need negative examples, run through the same embedding
        j_input = k.Input(shape=(1,), dtype='int32', name='neg-item')
        j_flat = k.layers.Flatten(name='neg-vector')(i_embed(j_input))

        # Score positive items with the dot product
        score = k.layers.Dot(name='pos-score', axes=1)([u_flat, i_flat])
        # And score negative items too
        neg_score = k.layers.Dot(name='neg-score', axes=1)([u_flat, j_flat])
        # Training is based on score differences
        train_score = k.layers.Subtract(name='score-diff')([score, neg_score])

        # Assemble the model for prediction
        model = k.Model([u_input, i_input], score, name='bpr-mf')
        # Assemble the training model and configure to optimize
        train = k.Model([u_input, i_input, j_input], train_score, name='bpr-train')
        train.compile('adam', BprLoss())

        return train, model

    def predict_for_user(self, user, items, ratings=None):
        if user not in self.user_index_:
            return pd.Series(np.nan, index=items)

        items = np.array(items)
        uno = self.user_index_.get_loc(user)
        inos = self.item_index_.get_indexer_for(items).astype('i4')
        good_inos = inos[inos >= 0]
        good_items = items[inos >= 0]
        unos = np.full(len(good_inos), uno, dtype='i4')
        _log.debug('scoring %d items for user %d', len(good_inos), user)

        ys = self.model.predict([unos, good_inos])

        res = pd.Series(ys[:, 0], index=good_items)
        return res.reindex(items)

    def __getstate__(self):
        state = dict(self.__dict__)
        if self.model is not None:
            # we need to save the model
            del state['model']
            _log.info('extracting model config and weights')
            state['model_config'] = self.model.get_config()
            state['model_weights'] = self.model.get_weights()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if 'model_config' in state:
            _log.info('rehydrating model')
            self.model = k.Model.from_config(state['model_config'])
            self.model.set_weights(state['model_weights'])
            del self.model_config
            del self.model_weights



"""
Prediction accuracy metrics.
"""

import numpy as np
import pandas as pd


def _check_missing(truth, missing):
    """
    Check for missing truth values.

    Args:
        truth: the series of truth values
        missing: what to do with missing values
    """
    if missing == 'error' and truth.isna().any():
        nmissing = truth.isna().sum()
        raise ValueError('missing truth for {} predictions'.format(nmissing))


def rmse(predictions, truth, missing='error'):
    """
    Compute RMSE (root mean squared error).

    Args:
        predictions(pandas.Series): the predictions
        truth(pandas.Series): the ground truth ratings from data
        missing(string):
            how to handle predictions without truth. Can be one of
            ``'error'`` or ``'ignore'``.

    Returns:
        double: the root mean squared approximation error
    """

    # force into series (basically no-op if already a series)
    predictions = pd.Series(predictions)
    truth = pd.Series(truth)

    # realign
    predictions, truth = predictions.align(truth, join='left')
    _check_missing(truth, missing)

    diff = predictions - truth

    sqdiff = diff.apply(np.square)
    msq = sqdiff.mean()
    return np.sqrt(msq)


def mae(predictions, truth, missing='error'):
    """
    Compute MAE (mean absolute error).

    Args:
        predictions(pandas.Series): the predictions
        truth(pandas.Series): the ground truth ratings from data
        missing(string):
            how to handle predictions without truth. Can be one of
            ``'error'`` or ``'ignore'``.

    Returns:
        double: the mean absolute approximation error
    """

    # force into series
    predictions = pd.Series(predictions)
    truth = pd.Series(truth)

    predictions, truth = predictions.align(truth, join='left')
    _check_missing(truth, missing)

    diff = predictions - truth

    adiff = diff.apply(np.abs)
    return adiff.mean()



import logging
import warnings

import pandas as pd

from .. import util

_logger = logging.getLogger(__name__)
_rec_context = None


def _predict_user(model, req):
    user, udf = req
    watch = util.Stopwatch()
    res = model.predict_for_user(user, udf['item'])
    res = pd.DataFrame({'user': user, 'item': res.index, 'prediction': res.values})
    _logger.debug('%s produced %f/%d predictions for %s in %s',
                  model, res.prediction.notna().sum(), len(udf), user, watch)
    return res


def predict(algo, pairs, *, n_jobs=None, **kwargs):
    """
    Generate predictions for user-item pairs.  The provided algorithm should be a
    :py:class:`algorithms.Predictor` or a function of two arguments: the user ID and
    a list of item IDs. It should return a dictionary or a :py:class:`pandas.Series`
    mapping item IDs to predictions.

    To use this function, provide a pre-fit algorithm::

        >>> from lenskit.algorithms.bias import Bias
        >>> from lenskit.metrics.predict import rmse
        >>> from lenskit import datasets
        >>> ratings = datasets.MovieLens('data/ml-latest-small').ratings
        >>> bias = Bias()
        >>> bias.fit(ratings[:-1000])
        <lenskit.algorithms.bias.Bias object at ...>
        >>> preds = predict(bias, ratings[-1000:])
        >>> preds.head()
               user  item  rating   timestamp  prediction
        99004   664  8361     3.0  1393891425    3.288286
        99005   664  8528     3.5  1393891047    3.559119
        99006   664  8529     4.0  1393891173    3.573008
        99007   664  8636     4.0  1393891175    3.846268
        99008   664  8641     4.5  1393890852    3.710635
        >>> rmse(preds['prediction'], preds['rating'])
        0.8326992222...

    Args:
        algo(lenskit.algorithms.Predictor):
            A rating predictor function or algorithm.
        pairs(pandas.DataFrame):
            A data frame of (``user``, ``item``) pairs to predict for. If this frame also
            contains a ``rating`` column, it will be included in the result.
        n_jobs(int):
            The number of processes to use for parallel batch prediction.  Passed to
            :func:`lenskit.util.parallel.invoker`.

    Returns:
        pandas.DataFrame:
            a frame with columns ``user``, ``item``, and ``prediction`` containing
            the prediction results. If ``pairs`` contains a `rating` column, this
            result will also contain a `rating` column.
    """
    if n_jobs is None and 'nprocs' in kwargs:
        n_jobs = kwargs['nprocs']
        warnings.warn('nprocs is deprecated, use n_jobs', DeprecationWarning)

    nusers = pairs['user'].nunique()

    timer = util.Stopwatch()
    with util.parallel.invoker(algo, _predict_user, n_jobs=n_jobs) as worker:
        del algo  # maybe free some memory

        _logger.info('generating %d predictions for %d users (setup took %s)',
                     len(pairs), nusers, timer)
        timer = util.Stopwatch()
        results = worker.map((user, udf.copy()) for (user, udf) in pairs.groupby('user'))
        results = pd.concat(results)
        _logger.info('generated %d predictions for %d users in %s',
                     len(pairs), nusers, timer)

    if 'rating' in pairs:
        return pairs.join(results.set_index(['user', 'item']), on=('user', 'item'))
    return results



"""
Data utilities
"""

import logging
import pathlib

try:
    import fastparquet
except ImportError:
    fastparquet = None

_log = logging.getLogger(__name__)


def read_df_detect(path):
    """
    Read a Pandas data frame, auto-detecting the file format based on filename suffix.
    The following file types are supported:

    CSV
        File has suffix ``.csv``, read with :py:func:`pandas.read_csv`.
    Parquet
        File has suffix ``.parquet``, ``.parq``, or ``.pq``, read with
        :py:func:`pandas.read_parquet`.
    """
    import pandas as pd
    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)

    if path.suffix == '.csv':
        return pd.read_csv(path)
    elif path.suffix in ('.parquet', '.parq', '.pq'):
        return pd.read_parquet(path)



"""
Item-based k-NN collaborative filtering.
"""

from sys import intern
import logging
import warnings

import pandas as pd
import numpy as np
import scipy.sparse as sps
import scipy.sparse.linalg as spla
from numba import jit, njit, prange
from numba.typed import List

from lenskit import util, matrix, DataWarning
from lenskit.sharing import in_share_context
from lenskit.util.parallel import is_mp_worker
from lenskit.util.accum import kvp_minheap_insert, kvp_minheap_sort
from . import Predictor

_logger = logging.getLogger(__name__)
_mkl_ops = matrix.mkl_ops()
_empty_csr = matrix._impl_mod()._empty_csr
_subset_rows = matrix._impl_mod()._subset_rows

if _mkl_ops is not None:
    # we have to import LK CFFI utils into this module
    for lkv in dir(_mkl_ops):
        if lkv.startswith('_lk_mkl'):
            globals()[lkv] = getattr(_mkl_ops, lkv)


def _make_blocks(n, size):
    "Create blocks for the range 0..n."
    return [(s, min(s + size, n)) for s in range(0, n, size)]


@njit(nogil=True)
def _count_nbrs(mat: matrix._CSR, thresh: float):
    "Count the number of neighbors passing the threshold for each row."
    counts = np.zeros(mat.nrows, dtype=np.int32)
    cs = mat.colinds
    vs = mat.values
    for i in range(mat.nrows):
        sp, ep = mat.row_extent(i)
        for j in range(sp, ep):
            c = cs[j]
            v = vs[j]
            if c != i and v >= thresh:
                counts[i] += 1

    return counts


@njit
def _insert(dst, used, limits, i, c, v):
    "Insert one item into a heap"
    sp = dst.rowptrs[i]
    ep = sp + used[i]
    ep = kvp_minheap_insert(sp, ep, limits[i], c, v, dst.colinds, dst.values)
    used[i] = ep - sp


@njit(nogil=True)
def _copy_nbrs(src: matrix._CSR, dst: matrix._CSR, limits, thresh: float):
    "Copy neighbors into the output matrix."
    used = np.zeros(dst.nrows, dtype=np.int32)

    for i in range(src.nrows):
        sp, ep = src.row_extent(i)

        for j in range(sp, ep):
            c = src.colinds[j]
            v = src.values[j]
            if c != i and v >= thresh:
                _insert(dst, used, limits, i, c, v)

    return used


@njit(nogil=True, parallel=not is_mp_worker())
def _sort_nbrs(smat):
    for i in prange(smat.nrows):
        sp, ep = smat.row_extent(i)
        kvp_minheap_sort(sp, ep, smat.colinds, smat.values)


@njit(nogil=True)
def _make_sim_block(nitems, bsp, bitems, r_sp, r_ep, r_cs, r_vs, min_sim, max_nbrs):
    # pass 1: compute the size of each row
    sizes = np.zeros(bitems, np.int32)
    for i in range(nitems):
        for j in range(r_sp[i], r_ep[i]):
            # we accept the neighbor if it passes threshold and isn't a self-similarity
            r = r_cs[j]
            if i != bsp + r and r_vs[j] >= min_sim:
                sizes[r] += 1

    if max_nbrs > 0:
        for i in range(bitems):
            if sizes[i] > max_nbrs:
                sizes[i] = max_nbrs

    # if bnc == 0:
    #     # empty resulting matrix, oops
    #     return _empty_csr(bitems, nitems, np.zeros(bitems, np.int32))

    # allocate a matrix
    block_csr = _empty_csr(bitems, nitems, sizes)

    # pass 2: truncate each row into the matrix
    eps = block_csr.rowptrs[:-1].copy()
    for c in range(nitems):
        for j in range(r_sp[c], r_ep[c]):
            v = r_vs[j]
            r = r_cs[j]
            sp, lep = block_csr.row_extent(r)
            lim = lep - sp
            if c != bsp + r and v >= min_sim:
                eps[r] = kvp_minheap_insert(sp, eps[r], lim, c, v,
                                            block_csr.colinds, block_csr.values)
        # we're done!
    return block_csr


@njit(nogil=True)
def _mkl_sim_block(block, bsp, bep, rmh, min_sim, max_nbrs, nitems):
    "Compute a single block of the similarity matrix"
    # assert block.nrows == bep - bsp

    bitems = block.nrows
    if block.nnz == 0:
        return _empty_csr(bitems, nitems, np.zeros(bitems, np.int32))

    # create a matrix handle for the subset matrix
    amh = _mkl_ops._from_csr(block)
    _lk_mkl_spopt(amh)

    smh = _lk_mkl_spmabt(rmh, amh)

    _lk_mkl_spfree(amh)

    _lk_mkl_sporder(smh)  # for reproducibility

    block = _lk_mkl_spexport_p(smh)
    bnr = _lk_mkl_spe_nrows(block)
    bnc = _lk_mkl_spe_ncols(block)
    # bnr and bnc should be right
    # assert bnc == bep - bsp
    # assert bnr == nitems

    r_sp = _lk_mkl_spe_row_sp(block)
    r_ep = _lk_mkl_spe_row_ep(block)
    r_cs = _lk_mkl_spe_colinds(block)
    r_vs = _lk_mkl_spe_values(block)

    block_csr = _make_sim_block(nitems, bsp, bitems, r_sp, r_ep, r_cs, r_vs, min_sim, max_nbrs)

    _lk_mkl_spe_free(block)
    _lk_mkl_spfree(smh)
    return block_csr


@njit(nogil=True, parallel=not is_mp_worker())
def _mkl_sim_blocks(trmat, blocks, ptrs, min_sim, max_nbrs):
    "Compute the similarity matrix with blocked MKL calls"
    nitems = trmat.nrows
    nblocks = len(blocks)

    null = _empty_csr(1, 1, np.zeros(1, dtype=np.int32))
    res = [null for i in range(nblocks)]

    rmat_h = _mkl_ops._from_csr(trmat)
    _lk_mkl_sporder(rmat_h)
    _lk_mkl_spopt(rmat_h)

    for bi in prange(nblocks):
        b = blocks[bi]
        p = ptrs[bi]
        bs, be = p
        bres = _mkl_sim_block(b, bs, be, rmat_h, min_sim, max_nbrs, nitems)
        res[bi] = bres

    _lk_mkl_spfree(rmat_h)

    return res


def _scipy_sim_block(block, bsp, bep, rmat, min_sim, max_nbrs, nitems):
    "Compute a single block of the similarity matrix"
    assert block.nrows == bep - bsp

    _logger.debug('processing block %d:%d (%d nnz)', bsp, bep, block.nnz)

    if rmat.nnz == 0:
        return _empty_csr(block.nrows, nitems, np.zeros(block.nrows, np.int32))

    sims = rmat @ block.to_scipy().transpose()
    sims = matrix.CSR.from_scipy(sims)

    r_sp = sims.rowptrs[:-1]
    r_ep = sims.rowptrs[1:]
    r_cs = sims.colinds
    r_vs = sims.values

    block_csr = _make_sim_block(nitems, bsp, block.nrows, r_sp, r_ep, r_cs, r_vs, min_sim, max_nbrs)
    assert block_csr.nrows == block.nrows
    assert block_csr.ncols == nitems
    _logger.debug('block %d:%d has %d similarities', bsp, bep, block_csr.nnz)
    _logger.debug('block: %s', matrix.CSR(N=block_csr))

    return block_csr


# we compile this in object mode so we can use numba's thread pool with scipy
@jit(parallel=not is_mp_worker(), forceobj=True)
def _scipy_sim_blocks(trmat, blocks, ptrs, min_sim, max_nbrs):
    "Compute the similarity matrix with blocked SciPy calls"
    nitems, nusers = trmat.shape
    nblocks = len(blocks)

    null = _empty_csr(1, 1, np.zeros(1, dtype=np.int32))
    res = [null for i in range(nblocks)]

    for bi in prange(nblocks):
        b = blocks[bi]
        bs, be = ptrs[bi]
        bres = _scipy_sim_block(b, bs, be, trmat, min_sim, max_nbrs, nitems)
        res[bi] = bres
        assert bres.nrows == be - bs

    return res


@njit(nogil=True)
def _predict_weighted_average(model, nitems, nrange, ratings, rated, targets):
    "Weighted average prediction function"
    min_nbrs, max_nbrs = nrange
    scores = np.full(nitems, np.nan, dtype=np.float_)

    for i in prange(targets.shape[0]):
        iidx = targets[i]
        rptr = model.rowptrs[iidx]
        rend = model.rowptrs[iidx + 1]

        num = 0
        denom = 0
        nnbrs = 0

        for j in range(rptr, rend):
            nidx = model.colinds[j]
            if not rated[nidx]:
                continue

            nnbrs = nnbrs + 1
            num = num + ratings[nidx] * model.values[j]
            denom = denom + np.abs(model.values[j])

            if max_nbrs > 0 and nnbrs >= max_nbrs:
                break

        if nnbrs < min_nbrs:
            continue

        scores[iidx] = num / denom

    return scores


@njit(nogil=True)
def _predict_sum(model, nitems, nrange, ratings, rated, targets):
    "Sum-of-similarities prediction function"
    min_nbrs, max_nbrs = nrange
    scores = np.full(nitems, np.nan, dtype=np.float_)

    for i in prange(targets.shape[0]):
        iidx = targets[i]
        rptr = model.rowptrs[iidx]
        rend = model.rowptrs[iidx + 1]

        score = 0
        nnbrs = 0

        for j in range(rptr, rend):
            nidx = model.colinds[j]
            if not rated[nidx]:
                continue

            nnbrs = nnbrs + 1
            score = score + model.values[j]

            if max_nbrs > 0 and nnbrs >= max_nbrs:
                break

        if nnbrs < min_nbrs:
            continue

        scores[iidx] = score

    return scores


_predictors = {
    'weighted-average': _predict_weighted_average,
    'sum': _predict_sum
}


class ItemItem(Predictor):
    """
    Item-item nearest-neighbor collaborative filtering with ratings. This item-item implementation
    is not terribly configurable; it hard-codes design decisions found to work well in the previous
    Java-based LensKit code.

    Args:
        nnbrs(int):
            the maximum number of neighbors for scoring each item (``None`` for unlimited)
        min_nbrs(int): the minimum number of neighbors for scoring each item
        min_sim(double): minimum similarity threshold for considering a neighbor
        save_nbrs(double):
            the number of neighbors to save per item in the trained model
            (``None`` for unlimited)
        center(bool):
            whether to normalize (mean-center) rating vectors.  Turn this off when working
            with unary data and other data types that don't respond well to centering.
        aggregate:
            the type of aggregation to do. Can be ``weighted-average`` or ``sum``.

    Attributes:
        item_index_(pandas.Index): the index of item IDs.
        item_means_(numpy.ndarray): the mean rating for each known item.
        item_counts_(numpy.ndarray): the number of saved neighbors for each item.
        sim_matrix_(matrix.CSR): the similarity matrix.
        user_index_(pandas.Index): the index of known user IDs for the rating matrix.
        rating_matrix_(matrix.CSR): the user-item rating matrix for looking up users' ratings.
    """
    AGG_SUM = intern('sum')
    AGG_WA = intern('weighted-average')
    _use_mkl = True

    def __init__(self, nnbrs, min_nbrs=1, min_sim=1.0e-6, save_nbrs=None,
                 center=True, aggregate='weighted-average'):
        self.nnbrs = nnbrs
        if self.nnbrs is not None and self.nnbrs < 1:
            self.nnbrs = -1
        self.min_nbrs = min_nbrs
        if self.min_nbrs is not None and self.min_nbrs < 1:
            self.min_nbrs = 1
        self.min_sim = min_sim
        self.save_nbrs = save_nbrs
        self.center = center
        self.aggregate = aggregate

    def fit(self, ratings, **kwargs):
        """
        Train a model.

        The model-training process depends on ``save_nbrs`` and ``min_sim``, but *not* on other
        algorithm parameters.

        Args:
            ratings(pandas.DataFrame):
                (user,item,rating) data for computing item similarities.
        """
        # Training proceeds in 2 steps:
        # 1. Normalize item vectors to be mean-centered and unit-normalized
        # 2. Compute similarities with pairwise dot products
        self._timer = util.Stopwatch()

        _logger.debug('[%s] beginning fit, memory use %s', self._timer, util.max_memory())

        init_rmat, users, items = matrix.sparse_ratings(ratings)
        n_items = len(items)
        _logger.info('[%s] made sparse matrix for %d items (%d ratings from %d users)',
                     self._timer, len(items), init_rmat.nnz, len(users))
        _logger.debug('[%s] made matrix, memory use %s', self._timer, util.max_memory())

        rmat, item_means = self._mean_center(ratings, init_rmat, items)
        _logger.debug('[%s] centered, memory use %s', self._timer, util.max_memory())

        rmat = self._normalize(rmat)
        _logger.debug('[%s] normalized, memory use %s', self._timer, util.max_memory())

        _logger.info('[%s] computing similarity matrix', self._timer)
        smat = self._compute_similarities(rmat)
        _logger.debug('[%s] computed, memory use %s', self._timer, util.max_memory())

        _logger.info('[%s] got neighborhoods for %d of %d items',
                     self._timer, np.sum(np.diff(smat.rowptrs) > 0), n_items)

        _logger.info('[%s] computed %d neighbor pairs', self._timer, smat.nnz)

        self.item_index_ = items
        self.item_means_ = item_means
        self.item_counts_ = np.diff(smat.rowptrs)
        self.sim_matrix_ = smat
        self.user_index_ = users
        self.rating_matrix_ = init_rmat
        # create an inverted similarity matrix for efficient scanning
        self._sim_inv_ = smat.transpose()
        _logger.info('[%s] transposed matrix for optimization', self._timer)
        _logger.debug('[%s] done, memory use %s', self._timer, util.max_memory())

        return self

    def _mean_center(self, ratings, rmat, items):
        if not self.center:
            return rmat, None

        item_means = ratings.groupby('item').rating.mean()
        item_means = item_means.reindex(items).values
        mcvals = rmat.values - item_means[rmat.colinds]
        nmat = matrix.CSR(rmat.nrows, rmat.ncols, rmat.nnz,
                          rmat.rowptrs.copy(), rmat.colinds.copy(), mcvals)
        _logger.info('[%s] computed means for %d items', self._timer, len(item_means))
        return nmat, item_means

    def _normalize(self, rmat):
        rmat = rmat.to_scipy()
        # compute column norms
        norms = spla.norm(rmat, 2, axis=0)
        # and multiply by a diagonal to normalize columns
        recip_norms = norms.copy()
        is_nz = recip_norms > 0
        recip_norms[is_nz] = np.reciprocal(recip_norms[is_nz])
        norm_mat = rmat @ sps.diags(recip_norms)
        assert norm_mat.shape[1] == rmat.shape[1]
        # and reset NaN
        norm_mat.data[np.isnan(norm_mat.data)] = 0
        _logger.info('[%s] normalized rating matrix columns', self._timer)
        return matrix.CSR.from_scipy(norm_mat, False)

    def _compute_similarities(self, rmat):
        trmat = rmat.transpose()
        nitems = trmat.nrows
        m_nbrs = self.save_nbrs
        if m_nbrs is None or m_nbrs < 0:
            m_nbrs = 0

        bounds = _make_blocks(nitems, 1000)
        _logger.info('[%s] splitting %d items (%d ratings) into %d blocks',
                     self._timer, nitems, trmat.nnz, len(bounds))
        blocks = [trmat.subset_rows(sp, ep) for (sp, ep) in bounds]

        if self._use_mkl and _mkl_ops is not None:
            _logger.info('[%s] computing similarities with MKL', self._timer)
            ptrs = List(bounds)
            nbs = List(b.N for b in blocks)
            if not nbs:
                # oops, this is the bad place
                # in non-JIT node, List doesn't actually make the list
                nbs = [b.N for b in blocks]
                ptrs = bounds
            s_blocks = _mkl_sim_blocks(trmat.N, nbs, ptrs, self.min_sim, m_nbrs)
        else:
            s_blocks = _scipy_sim_blocks(trmat.to_scipy(), blocks, bounds, self.min_sim, m_nbrs)

        s_blocks = [matrix.CSR(N=b) for b in s_blocks]
        nnz = sum(b.nnz for b in s_blocks)
        tot_rows = sum(b.nrows for b in s_blocks)
        _logger.info('[%s] computed %d similarities for %d items in %d blocks',
                     self._timer, nnz, tot_rows, len(s_blocks))
        row_nnzs = np.concatenate([b.row_nnzs() for b in s_blocks])
        assert len(row_nnzs) == nitems, \
            'only have {} rows for {} items'.format(len(row_nnzs), nitems)

        smat = matrix.CSR.empty((nitems, nitems), row_nnzs, rpdtype=np.int64)
        start = 0
        for bi, b in enumerate(s_blocks):
            bnr = b.nrows
            end = start + bnr
            v_sp = smat.rowptrs[start]
            v_ep = smat.rowptrs[end]
            _logger.debug('block %d (%d:%d) has %d entries, storing in %d:%d',
                          bi, start, end, b.nnz, v_sp, v_ep)
            smat.colinds[v_sp:v_ep] = b.colinds
            smat.values[v_sp:v_ep] = b.values
            start = end

        _logger.info('[%s] sorting similarity matrix with %d entries', self._timer, smat.nnz)
        _sort_nbrs(smat.N)

        return smat

    def predict_for_user(self, user, items, ratings=None):
        _logger.debug('predicting %d items for user %s', len(items), user)
        if ratings is None:
            if user not in self.user_index_:
                _logger.debug('user %s missing, returning empty predictions', user)
                return pd.Series(np.nan, index=items)
            upos = self.user_index_.get_loc(user)
            ratings = pd.Series(self.rating_matrix_.row_vs(upos),
                                index=pd.Index(self.item_index_[self.rating_matrix_.row_cs(upos)]))

        if not ratings.index.is_unique:
            wmsg = 'user {} has duplicate ratings, this is likely to cause problems'.format(user)
            warnings.warn(wmsg, DataWarning)

        # set up rating array
        # get rated item positions & limit to in-model items
        n_items = len(self.item_index_)
        ri_pos = self.item_index_.get_indexer(ratings.index)
        m_rates = ratings[ri_pos >= 0]
        ri_pos = ri_pos[ri_pos >= 0]
        rate_v = np.full(n_items, np.nan, dtype=np.float_)
        rated = np.zeros(n_items, dtype='bool')
        # mean-center the rating array
        if self.center:
            rate_v[ri_pos] = m_rates.values - self.item_means_[ri_pos]
        else:
            rate_v[ri_pos] = m_rates.values
        rated[ri_pos] = True

        _logger.debug('user %s: %d of %d rated items in model', user, len(ri_pos), len(ratings))
        assert np.sum(np.logical_not(np.isnan(rate_v))) == len(ri_pos)
        assert np.all(np.isnan(rate_v) == np.logical_not(rated))

        # set up item result vector
        # ipos will be an array of item indices
        i_pos = self.item_index_.get_indexer(items)
        i_pos = i_pos[i_pos >= 0]
        _logger.debug('user %s: %d of %d requested items in model', user, len(i_pos), len(items))

        # now we take a first pass through the data to count _viable_ targets
        # This computes the number of neighbors (and their weight sum) for
        # each target item based on the user's ratings, allowing us to fast-path
        # other computations and avoid as many neighbor truncations as possible
        i_cts, i_sums, i_nbrs = self._count_viable_targets(i_pos, ri_pos)
        viable = i_cts >= self.min_nbrs
        i_pos = i_pos[viable]
        i_cts = i_cts[viable]
        i_sums = i_sums[viable]
        i_nbrs = i_nbrs[viable]
        _logger.debug('user %s: %d of %d requested items possibly reachable',
                      user, len(i_pos), len(items))

        # look for some fast paths
        if self.aggregate == self.AGG_SUM and self.min_sim >= 0:
            # similarity sums are all we need
            if self.nnbrs >= 0:
                fast_mask = i_cts <= self.nnbrs
                fast_items = i_pos[fast_mask]
                fast_scores = i_sums[fast_mask]
                slow_items = i_pos[~fast_mask]
            else:
                fast_items = i_pos
                fast_scores = i_sums
                slow_items = np.array([], dtype='i4')

            _logger.debug('user %s: using fast-path similarity sum for %d items',
                          user, len(fast_items))

            if len(slow_items):
                iscores = _predict_sum(self.sim_matrix_.N, len(self.item_index_),
                                       (self.min_nbrs, self.nnbrs),
                                       rate_v, rated, slow_items)
            else:
                iscores = np.full(len(self.item_index_), np.nan)
            iscores[fast_items] = fast_scores

        elif self.aggregate == self.AGG_WA and self.min_nbrs == 1:
            # fast-path single-neighbor targets - common in sparse data
            fast_mask = i_cts == 1
            fast_items = i_pos[fast_mask]
            fast_scores = rate_v[i_nbrs[fast_mask]]
            if self.min_sim < 0:
                fast_scores *= np.sign(i_sums[fast_mask])
            _logger.debug('user %s: fast-pathed %d scores', user, len(fast_scores))

            slow_items = i_pos[i_cts > 1]
            iscores = _predict_weighted_average(self.sim_matrix_.N, len(self.item_index_),
                                                (self.min_nbrs, self.nnbrs),
                                                rate_v, rated, slow_items)
            iscores[fast_items] = fast_scores
        else:
            # now compute the predictions
            _logger.debug('user %s: taking the slow path', user)
            agg = _predictors[self.aggregate]
            iscores = agg(self.sim_matrix_.N, len(self.item_index_), (self.min_nbrs, self.nnbrs),
                          rate_v, rated, i_pos)

        if self.center:
            iscores += self.item_means_

        results = pd.Series(iscores, index=self.item_index_)
        results = results.reindex(items, fill_value=np.nan)

        _logger.debug('user %s: predicted for %d of %d items',
                      user, results.notna().sum(), len(items))

        return results

    def _count_viable_targets(self, targets, rated):
        "Count upper-bound on possible neighbors for target items and rated items."
        # initialize counts to zero
        counts = np.zeros(len(self.item_index_), dtype=np.int32)
        sums = np.zeros(len(self.item_index_))
        last_nbrs = np.full(len(self.item_index_), -1, 'i4')
        # count the number of times each item is reachable from the neighborhood
        for ri in rated:
            nbrs = self._sim_inv_.row_cs(ri)
            counts[nbrs] += 1
            sums[nbrs] += self._sim_inv_.row_vs(ri)
            last_nbrs[nbrs] = ri

        # we want the reachability counts for the target items
        return counts[targets], sums[targets], last_nbrs[targets]

    def __getstate__(self):
        state = dict(self.__dict__)
        if '_sim_inv_' in state and not in_share_context():
            del state['_sim_inv_']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if hasattr(self, 'sim_matrix_') and not hasattr(self, '_sim_inv_'):
            self._sim_inv_ = self.sim_matrix_.transpose()

    def __str__(self):
        return 'ItemItem(nnbrs={}, msize={})'.format(self.nnbrs, self.save_nbrs)



import os
import pathlib
import tempfile
import logging
import gc

import binpickle

from . import sharing_mode, PersistedModel

_log = logging.getLogger(__name__)


def persist_binpickle(model, dir=None, file=None):
    """
    Persist a model using binpickle.

    Args:
        model: The model to persist.
        dir: The temporary directory for persisting the model object.
        file: The file in which to save the object.

    Returns:
        PersistedModel: The persisted object.
    """
    if file is not None:
        path = pathlib.Path(file)
    else:
        if dir is None:
            dir = os.environ.get('LK_TEMP_DIR', None)
        fd, path = tempfile.mkstemp(suffix='.bpk', prefix='lkpy-', dir=dir)
        os.close(fd)
        path = pathlib.Path(path)
    _log.debug('persisting %s to %s', model, path)
    with binpickle.BinPickler.mappable(path) as bp, sharing_mode():
        bp.dump(model)
    return BPKPersisted(path)


class BPKPersisted(PersistedModel):
    def __init__(self, path):
        self.path = path
        self.is_owner = True
        self._bpk_file = None
        self._model = None

    def get(self):
        if self._bpk_file is None:
            _log.debug('loading %s', self.path)
            self._bpk_file = binpickle.BinPickleFile(self.path, direct=True)
            self._model = self._bpk_file.load()
        return self._model

    def close(self, unlink=True):
        if self._bpk_file is not None:
            self._model = None
            try:
                _log.debug('closing BPK file')
                try:
                    self._bpk_file.close()
                except BufferError:
                    _log.debug('could not close %s, collecting garbage and retrying', self.path)
                    gc.collect()
                    self._bpk_file.close()
            except (BufferError, IOError) as e:
                _log.warn('error closing %s: %s', self.path, e)
            self._bpk_file = None

        if self.is_owner and unlink:
            assert self._model is None
            if unlink:
                _log.debug('deleting %s', self.path)
                try:
                    self.path.unlink()
                except IOError as e:
                    _log.warn('could not remove %s: %s', self.path, e)
            self.is_owner = False

    def __getstate__(self):
        d = dict(self.__dict__)
        d['_bpk_file'] = None
        d['_model'] = None
        if self.is_owner == 'transfer':
            d['is_owner'] = True
        else:
            d['is_owner'] = False
        return d

    def __del___(self):
        self.close(False)



import os
import pathlib
import logging

import cffi
from distutils import ccompiler
from numba import njit, types as nt

try:
    from numba.core.typing import cffi_utils
except ImportError:
    # earlier Numba versions have an old module
    from numba import cffi_support as cffi_utils

import numpy as np

from . import CSR, _impl_mod
_CSR = _impl_mod()._CSR

_logger = logging.getLogger(__name__)
__dir = pathlib.Path(__file__).parent

__cc = ccompiler.new_compiler()
_mkl_so = __dir / __cc.shared_object_filename('mkl_ops')
__mkl_defs = (__dir / 'mkl_ops.h').read_text()
ffi = cffi.FFI()
ffi.cdef(__mkl_defs.replace('EXPORT ', ''))
try:
    clib = ffi.dlopen(os.fspath(_mkl_so))

    _lk_mkl_spcreate = clib.lk_mkl_spcreate
    _lk_mkl_spsubset = clib.lk_mkl_spsubset
    _lk_mkl_spfree = clib.lk_mkl_spfree
    _lk_mkl_sporder = clib.lk_mkl_sporder
    _lk_mkl_spopt = clib.lk_mkl_spopt
    _lk_mkl_spmv = clib.lk_mkl_spmv
    _lk_mkl_spmab = clib.lk_mkl_spmab
    _lk_mkl_spmabt = clib.lk_mkl_spmabt
    _lk_mkl_spexport = clib.lk_mkl_spexport

    # silly pointer interface
    _lk_mkl_spexport_p = clib.lk_mkl_spexport_p
    _lk_mkl_spe_free = clib.lk_mkl_spe_free
    _lk_mkl_spe_nrows = clib.lk_mkl_spe_nrows
    _lk_mkl_spe_ncols = clib.lk_mkl_spe_ncols
    _lk_mkl_spe_row_sp = clib.lk_mkl_spe_row_sp
    _lk_mkl_spe_row_ep = clib.lk_mkl_spe_row_ep
    _lk_mkl_spe_colinds = clib.lk_mkl_spe_colinds
    _lk_mkl_spe_values = clib.lk_mkl_spe_values
except OSError:
    clib = None

# support intptr_t
if cffi_utils is not None:
    # if we don't have working Numba, skip this
    cffi_utils.register_type(ffi.typeof('intptr_t'), nt.intp)

# extract sizes
_int_size = ffi.sizeof('int')
_dbl_size = ffi.sizeof('double')

_mkl_errors = [
    'SPARSE_STATUS_SUCCESS',
    'SPARSE_STATUS_NOT_INITIALIZED',
    'SPARSE_STATUS_ALLOC_FAILED',
    'SPARSE_STATUS_INVALID_VALUE',
    'SPARSE_STATUS_EXECUTION_FAILED',
    'SPARSE_STATUS_INTERNAL_ERROR',
    'SPARSE_STATUS_NOT_SUPPORTED'
]


def _mkl_check_return(rv, call='<unknown>'):
    if rv:
        if rv >= 0 and rv < len(_mkl_errors):
            desc = _mkl_errors[rv]
        else:
            desc = 'unknown'
        raise RuntimeError('MKL call {} failed with code {} ({})'.format(call, rv, desc))


class SparseM:
    """
    Class encapsulating an MKL sparse matrix handle.
    """

    def __init__(self):
        self.ptr = None

    @classmethod
    def from_csr(cls, csr):
        """
        Create an MKL sparse matrix from a LensKit CSR matrix.

        Args:
            csr(CSR): the input matrix.

        Returns:
            SparseM: a sparse matrix handle for the CSR matrix.
        """
        m = SparseM()
        m.ptr = _from_csr(csr.N)
        if not m.ptr:
            raise RuntimeError('MKL matrix creation failed')

        m._csr = csr  # save the CSR matrix to ensure it oulives the SparseH
        return m

    def __del__(self):
        if self.ptr:
            _logger.debug('destroying MKL sparse matrix 0x%08x', id(self))
            clib.lk_mkl_spfree(self.ptr)

    def export(self):
        """
        Export an MKL sparse matrix as a LensKit CSR.

        Returns:
            CSR: the LensKit matrix.
        """
        csr = _to_csr(self.ptr)
        if not csr:
            raise RuntimeError('MKL failed to export CSR')

        return CSR(N=csr)

    def mult_vec(self, alpha, x, beta, y):
        """
        Compute :math:`\\alpha A x + \\beta y`, where :math:`A` is this matrix.
        """
        x = np.require(x, np.float64, 'C')
        yout = np.require(y, np.float64, 'C')
        if yout is y:
            yout = yout.copy()

        _x = ffi.from_buffer('double[]', x)
        _y = ffi.from_buffer('double[]', yout)

        rv = clib.lk_mkl_spmv(alpha, self.ptr, _x, beta, _y)
        _mkl_check_return(rv, 'mkl_sparse_d_mv')

        return yout


@njit
def _from_csr(csr: _CSR):
    """
    Convert a Numba CSR to an MKL sparse matrix handle.
    """
    _sp = ffi.from_buffer(csr.rowptrs)
    _cols = ffi.from_buffer(csr.colinds)
    _vals = ffi.from_buffer(csr.values)
    return _lk_mkl_spcreate(csr.nrows, csr.ncols, _sp, _cols, _vals)


@njit
def _from_csr_ss(csr: _CSR, rsp, rep):
    """
    Convert a subset of a Numba CSR to an MKL sparse matrix handle.
    """
    _sp = ffi.from_buffer(csr.rowptrs)
    _cols = ffi.from_buffer(csr.colinds)
    _vals = ffi.from_buffer(csr.values)
    return _lk_mkl_spsubset(rsp, rep, csr.ncols, _sp, _cols, _vals)


@njit
def _to_csr(smh):
    """
    Convert an MKL sparse matrix handle to a Numba CSR.
    """
    rvp = _lk_mkl_spexport_p(smh)
    if rvp is None:
        return None

    nrows = _lk_mkl_spe_nrows(rvp)
    ncols = _lk_mkl_spe_ncols(rvp)

    sp = _lk_mkl_spe_row_sp(rvp)
    ep = _lk_mkl_spe_row_ep(rvp)
    cis = _lk_mkl_spe_colinds(rvp)
    vs = _lk_mkl_spe_values(rvp)

    rowptrs = np.zeros(nrows + 1, dtype=np.intc)
    nnz = 0
    for i in range(nrows):
        nnz += ep[i] - sp[i]
        rowptrs[i+1] = nnz

    colinds = np.zeros(nnz, dtype=np.intc)
    values = np.zeros(nnz)

    for i in range(nrows):
        rs = rowptrs[i]
        re = rowptrs[i+1]
        ss = sp[i]
        for j in range(re - rs):
            colinds[rs + j] = cis[ss + j]
            values[rs + j] = vs[ss + j]

    _lk_mkl_spe_free(rvp)

    return _CSR(nrows, ncols, nnz, rowptrs, colinds, values)



"""
Batch-run predictors and recommenders for evaluation.
"""

from ._predict import predict  # noqa: F401
from ._recommend import recommend  # noqa: F401
from ._multi import MultiEval  # noqa: F401
from ._train import train_isolated  # noqa: F401



"""
Utilities to manage randomness in LensKit and LensKit experiments.
"""

import zlib
import numpy as np
import random
import warnings
import logging

_log = logging.getLogger(__name__)


class LegacyRNG:
    _seed = None
    _rng = None

    @property
    def seed(self):
        if self._seed is None:
            self._seed = np.random.randint(0, np.iinfo('i4').max)
        return self._seed

    @property
    def int_seed(self):
        return self._seed

    def initialize(self, seed, keys):
        warnings.warn('initializing legacy RNG infrastructure')
        _log.warn('initializing legacy RNG infrastructure - use NumPy 1.17+ for better seeds')
        if keys:
            raise NotImplementedError('legacy RNG does not support seed keys')
        self._seed = seed
        if self._rng is not None:
            del self._rng
        return seed

    def derive(self, base, keys):
        raise NotImplementedError('legacy RNG does not support deriving seeds')

    def rng(self, seed=None):
        if seed is None:
            if self._rng is None:
                seed = self.seed
                self._rng = np.random.RandomState(seed)
            return self._rng
        else:
            return np.random.RandomState(seed)


class ModernRNG:
    _seed = None

    @property
    def seed(self):
        if self._seed is None:
            self._seed = np.random.SeedSequence()
        return self._seed

    @property
    def int_seed(self):
        return self._seed.generate_state(1)[0]

    def initialize(self, seed, keys):
        if isinstance(seed, int):
            seed = np.random.SeedSequence(seed)

        if not isinstance(seed, np.random.SeedSequence):
            raise TypeError('unexpected seed type %s', type(seed))

        if keys:
            seed = self.derive(seed, keys)

        self._seed = seed
        return seed

    def derive(self, base, keys):
        if base is None:
            base = self.seed

        if keys:
            k2 = tuple(_make_int(k) for k in keys)
            return np.random.SeedSequence(base.entropy, spawn_key=base.spawn_key + k2)
        else:
            return base.spawn(1)[0]

    def rng(self, seed=None):
        if seed is None:
            seed, = self.seed.spawn(1)
        elif isinstance(seed, int):
            seed = np.random.SeedSequence(seed)
        return np.random.default_rng(seed)


def get_root_seed():
    """
    Get the root seed.

    Returns:
        numpy.random.SeedSequence: The LensKit root seed.
    """
    return _rng_impl.seed


def _make_int(obj):
    if isinstance(obj, int):
        return obj
    elif isinstance(obj, bytes):
        return zlib.crc32(obj)
    elif isinstance(obj, str):
        return zlib.crc32(obj.encode('utf8'))
    else:
        return ValueError('invalid RNG key ' + str(obj))


def init_rng(seed, *keys, propagate=True):
    """
    Initialize the random infrastructure with a seed.  This function should generally be
    called very early in the setup.

    Args:
        seed(int or numpy.random.SeedSequence):
            The random seed to initialize with.
        keys:
            Additional keys, to use as a ``spawn_key`` on NumPy 1.17.  Passed to
            :func:`derive_seed`.

        propagate(bool):
            If ``True``, initialize other RNG infrastructure. This currently initializes:

            * :func:`np.random.seed`
            * :func:`random.seed`

            If ``propagate=False``, LensKit is still fully seeded — no component included
            with LensKit uses any of the global RNGs, they all use RNGs seeded with the
            specified seed.

    Returns:
        The random seed.
    """
    _rng_impl.initialize(seed, keys)
    _log.info('initialized LensKit RNG with seed %s', _rng_impl.seed)

    if propagate:
        ik = _rng_impl.int_seed
        _log.info('initializing numpy.random and random with seed %u', ik)
        np.random.seed(ik)
        random.seed(ik)
        try:
            import tensorflow as tf
            _tf_seed = getattr(tf.random, 'set_seed', None)
            if _tf_seed is not None:
                _log.debug('setting TensorFlow seed')
                _tf_seed(ik)
            else:
                _log.warn('old TensorFlow, not setting seed')
        except ImportError:
            _log.debug('TensorFlow not available')

    return _rng_impl.seed


def derive_seed(*keys, base=None, none_on_old_numpy=False):
    """
    Derive a seed from the root seed, optionally with additional seed keys.

    Args:
        keys(list of int or str):
            Additional components to add to the spawn key for reproducible derivation.
            If unspecified, the seed's internal counter is incremented (by calling
            :meth:`numpy.random.SeedSequence.spawn`).
        base(numpy.random.SeedSequence):
            The base seed to use.  If ``None``, uses the root seed.
        none_on_old_numpy(bool):
            If ``True``, return ``None`` instead of raising :class:`NotImplementedError`
            if running on an old version of NumPy.
    """
    try:
        return _rng_impl.derive(base, keys)
    except NotImplementedError as e:
        if none_on_old_numpy:
            return None
        else:
            raise e


def rng_seed(spec=None):
    """
    Get a random number generator seed.  ``spec`` is interpreted as in :func:`rng`, and
    is used as follows:

    * If a :class:`numpy.random.SeedSequence`, returned as-is.
    * If an integer, used to create a seed sequence.
    * If ``None``, returns global seed (after initializing).
    * If a :class:`numpy.random.Generator` or :class:`numpy.random.RandomState`, it is
      used to generate an integer that is used to create a seed sequence.

    This function is only available when used with NumPy 1.17 or newer.

    Returns:
        numpy.random.SeedSequence:
            The seed.
    """
    if not hasattr(np.random, 'Generator'):
        raise RuntimeError('rng_seed requires NumPy 1.17')

    if spec is None:
        return _rng_impl.seed
    elif isinstance(spec, int):
        return np.random.SeedSequence(spec)
    elif isinstance(spec, np.random.SeedSequence):
        return spec
    elif hasattr(spec, 'integers'):
        seed = rng.integers(2**32-1)
        return np.random.SeedSequence(seed)
    else:
        raise ValueError('unknown RNG spec ' + str(spec))


def rng(spec=None, *, legacy=False):
    """
    Get a random number generator.  This is similar to :func:`sklearn.utils.check_random_seed`, but
    it usually returns a :class:`numpy.random.Generator` instead.

    Args:
        spec:
            The spec for this RNG.  Can be any of the following types:

            * ``int``
            * ``None``
            * :class:`numpy.random.SeedSequence`
            * :class:`numpy.random.mtrand.RandomState`
            * :class:`numpy.random.Generator`
        legacy(bool):
            If ``True``, return :class:`numpy.random.mtrand.RandomState` instead of a new-style
            :class:`numpy.random.Generator`.

    Returns:
        numpy.random.Generator: A random number generator.
    """

    rng = None
    if isinstance(spec, np.random.RandomState):
        rng = spec
    elif _have_gen and isinstance(spec, np.random.Generator):
        rng = spec
    else:
        rng = _rng_impl.rng(spec)

    if legacy and _have_gen and isinstance(rng, np.random.Generator):
        rng = np.random.RandomState(rng.bit_generator)
    # case where rng is a random state, and we are on new numpy and want a generator, is ok

    return rng


class FixedRNG:
    "RNG provider that always provides the same RNG"
    def __init__(self, rng):
        self.rng = rng

    def __call__(self, *keys):
        return self.rng

    def __str__(self):
        return 'Fixed({})'.format(self.rng)


class DerivingRNG:
    "RNG provider that derives new RNGs from the key"
    def __init__(self, seed, legacy):
        self.seed = seed
        self.legacy = legacy

    def __call__(self, *keys):
        seed = derive_seed(*keys, base=self.seed)
        if self.legacy:
            bg = np.random.MT19937(seed)
            return np.random.RandomState(bg)
        else:
            return np.random.default_rng(seed)

    def __str__(self):
        return 'Derive({})'.format(self.seed)


def derivable_rng(spec, *, legacy=False):
    """
    Get a derivable RNG, for use cases where the code needs to be able to reproducibly derive
    sub-RNGs for different keys, such as user IDs.

    Args:
        spec:
            Any value supported by the `seed` parameter of :func:`rng`, in addition to the
            following values:

            * the string ``'user'``
            * a tuple of the form (`seed`, ``'user'``)

            Either of these forms will cause the returned function to re-derive new RNGs.

    Returns:
        function:
            A function taking one (or more) key values, like :func:`derive_seed`, and
            returning a random number generator (the type of which is determined by
            the ``legacy`` parameter).
    """

    if spec == 'user':
        return DerivingRNG(derive_seed(), legacy)
    elif isinstance(spec, tuple):
        seed, key = spec
        if key != 'user':
            raise ValueError('unrecognized key %s', key)
        return DerivingRNG(seed, legacy)
    else:
        return FixedRNG(rng(spec, legacy=legacy))


# are we on modern NumPy?
_have_gen = hasattr(np.random, 'Generator')
if _have_gen:
    _rng_impl = ModernRNG()
else:
    _rng_impl = LegacyRNG()



"""
Utilities for parallel processing.
"""

import os
import multiprocessing as mp
from multiprocessing.queues import SimpleQueue
import functools as ft
import logging
import logging.handlers
import inspect
import faulthandler
from concurrent.futures import ProcessPoolExecutor
from abc import ABC, abstractmethod
import pickle

from lenskit.sharing import persist, PersistedModel
from lenskit.util.log import log_queue
from lenskit.util.random import derive_seed, init_rng, get_root_seed

if pickle.HIGHEST_PROTOCOL < 5:
    import pickle5 as pickle

_log = logging.getLogger(__name__)
__work_model = None
__work_func = None
__is_worker = False


def is_worker():
    "Query whether the process is a worker, either for MP or for isolation."
    return __is_worker


def is_mp_worker():
    "Query whether the current process is a multiprocessing worker."
    return os.environ.get('_LK_IN_MP', 'no') == 'yes'


def _p5_recv(self):
    buf = self.recv_bytes()
    return pickle.loads(buf)


def _p5_send(self, obj):
    buf = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
    self._send_bytes(buf)


class FastQ(SimpleQueue):
    """
    SimpleQueue subclass that uses Pickle5 instead of default pickling.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__patch()

    def __patch(self):
        # monkey-patch the sockets to use pickle5
        self._reader.recv = _p5_recv.__get__(self._reader)
        self._writer.send = _p5_send.__get__(self._writer)

    def get(self):
        with self._rlock:
            res = self._reader.recv_bytes()
        return pickle.loads(res)

    def put(self, obj):
        bytes = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
        # follow SimpleQueue, need to deal with _wlock being None
        if self._wlock is None:
            self._writer.send_bytes(bytes)
        else:
            with self._wlock:
                self._writer.send_bytes(bytes)

    def __setstate__(self, state):
        super().__setstate__(state)
        self.__patch()


class LKContext(mp.context.SpawnContext):
    def SimpleQueue(self):
        return FastQ(ctx=self.get_context())


LKContext.INSTANCE = LKContext()


def _initialize_worker(log_queue, seed):
    "Initialize a worker process."
    global __is_worker
    __is_worker = True
    faulthandler.enable()
    if seed is not None:
        init_rng(seed)
    if log_queue is not None:
        h = logging.handlers.QueueHandler(log_queue)
        root = logging.getLogger()
        root.addHandler(h)
        root.setLevel(logging.DEBUG)
        h.setLevel(logging.DEBUG)


def _initialize_mp_worker(mkey, func, threads, log_queue, seed):
    seed = derive_seed(mp.current_process().name, base=seed, none_on_old_numpy=True)
    _initialize_worker(log_queue, seed)
    global __work_model, __work_func

    nnt_env = os.environ.get('NUMBA_NUM_THREADS', None)
    if nnt_env is None or int(nnt_env) > threads:
        _log.debug('configuring Numba thread count')
        import numba
        numba.config.NUMBA_NUM_THREADS = threads
    try:
        import mkl
        _log.debug('configuring MKL thread count')
        mkl.set_num_threads(threads)
    except ImportError:
        pass

    __work_model = mkey
    # deferred function unpickling to minimize imports before initialization
    __work_func = pickle.loads(func)

    _log.debug('worker %d ready (process %s)', os.getpid(), mp.current_process())


def _mp_invoke_worker(*args):
    model = __work_model.get()
    return __work_func(model, *args)


def _sp_worker(log_queue, seed, res_queue, func, args, kwargs):
    _initialize_worker(log_queue, seed)
    _log.debug('running %s in worker', func)
    try:
        res = func(*args, **kwargs)
        _log.debug('completed successfully')
        res_queue.put((True, res))
    except Exception as e:
        _log.error('failed, transmitting error %r', e)
        res_queue.put((False, e))


def proc_count(core_div=2, max_default=None, level=0):
    """
    Get the number of desired jobs for multiprocessing operations.  This does not
    affect Numba or MKL multithreading.

    This count can come from a number of sources:

    * The ``LK_NUM_PROCS`` environment variable
    * The number of CPUs, divided by ``core_div`` (default 2)

    Args:
        core_div(int or None):
            The divisor to scale down the number of cores; ``None`` to turn off core-based
            fallback.
        max_default:
            The maximum number of processes to use if the environment variable is not
            configured.
        level:
            The process nesting level.  0 is the outermost level of parallelism; subsequent
            levels control nesting.  Levels deeper than 1 are rare, and it isn't expected
            that callers actually have an accurate idea of the threading nesting, just that
            they are configuring a child.  If the process count is unconfigured, then level
            1 will use ``core_div``, and deeper levels will use 1.

    Returns:
        int: The number of jobs desired.
    """

    nprocs = os.environ.get('LK_NUM_PROCS', None)
    if nprocs is not None:
        nprocs = [int(s) for s in nprocs.split(',')]
    elif core_div is not None:
        nprocs = max(mp.cpu_count() // core_div, 1)
        if max_default is not None:
            nprocs = min(nprocs, max_default)
        nprocs = [nprocs, core_div]

    if level >= len(nprocs):
        return 1
    else:
        return nprocs[level]


def run_sp(func, *args, **kwargs):
    """
    Run a function in a subprocess and return its value.  This is for achieving subprocess
    isolation, not parallelism.  The subprocess is configured so things like logging work
    correctly, and is initialized with a derived random seed.
    """
    ctx = LKContext.INSTANCE
    rq = ctx.SimpleQueue()
    seed = derive_seed(none_on_old_numpy=True)
    worker_args = (log_queue(), seed, rq, func, args, kwargs)
    _log.debug('spawning subprocess to run %s', func)
    proc = ctx.Process(target=_sp_worker, args=worker_args)
    proc.start()
    _log.debug('waiting for process %s to return', proc)
    success, payload = rq.get()
    _log.debug('received success=%s', success)
    _log.debug('waiting for process %s to exit', proc)
    proc.join()
    if proc.exitcode:
        _log.error('subprocess failed with code %d', proc.exitcode)
        raise RuntimeError('subprocess failed with code ' + str(proc.exitcode))
    if success:
        return payload
    else:
        _log.error('subprocess raised exception: %s', payload)
        raise ChildProcessError('error in child process', payload)


def invoker(model, func, n_jobs=None, *, persist_method=None):
    """
    Get an appropriate invoker for performing oeprations on ``model``.

    Args:
        model(obj): The model object on which to perform operations.
        func(function): The function to call.  The function must be pickleable.
        n_jobs(int or None):
            The number of processes to use for parallel operations.  If ``None``, will
            call :func:`proc_count` with a maximum default process count of 4.
        persist_method(str or None):
            The persistence method to use.  Passed as ``method`` to
            :func:`lenskit.sharing.persist`.

    Returns:
        ModelOpInvoker:
            An invoker to perform operations on the model.
    """
    if n_jobs is None:
        n_jobs = proc_count(max_default=4)

    if n_jobs == 1:
        return InProcessOpInvoker(model, func)
    elif 'mp_context' in inspect.signature(ProcessPoolExecutor).parameters:
        return ProcessPoolOpInvoker(model, func, n_jobs, persist_method)
    else:
        _log.warn('using multiprocessing.Pool, upgrade to Python 3.7 for best results')
        return MPOpInvoker(model, func, n_jobs, persist_method)


class ModelOpInvoker(ABC):
    """
    Interface for invoking operations on a model, possibly in parallel.  The operation
    invoker is configured with a model and a function to apply, and applies that function
    to the arguments supplied in `map`.  Child process invokers also route logging messages
    to the parent process, so logging works even with multiprocessing.

    An invoker is a context manager that calls :meth:`shutdown` when exited.
    """

    @abstractmethod
    def map(self, *iterables):
        """
        Apply the configured function to the model and iterables.  This is like :py:func:`map`,
        except it supplies the invoker's model as the first object to ``func``.

        Args:
            iterables: Iterables of arguments to provide to the function.

        Returns:
            iterable: An iterable of the results.
        """
        pass

    def shutdown(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.shutdown()


class InProcessOpInvoker(ModelOpInvoker):
    def __init__(self, model, func):
        _log.info('setting up in-process worker')
        if isinstance(model, PersistedModel):
            self.model = model.get()
        else:
            self.model = model
        self.function = func

    def map(self, *iterables):
        proc = ft.partial(self.function, self.model)
        return map(proc, *iterables)

    def shutdown(self):
        self.model = None


class ProcessPoolOpInvoker(ModelOpInvoker):
    def __init__(self, model, func, n_jobs, persist_method):
        if isinstance(model, PersistedModel):
            key = model
        else:
            key = persist(model, method=persist_method)
        func = pickle.dumps(func)
        ctx = LKContext.INSTANCE
        _log.info('setting up ProcessPoolExecutor w/ %d workers', n_jobs)
        os.environ['_LK_IN_MP'] = 'yes'
        kid_tc = proc_count(level=1)
        self.executor = ProcessPoolExecutor(n_jobs, ctx, _initialize_mp_worker,
                                            (key, func, kid_tc, log_queue(), get_root_seed()))

    def map(self, *iterables):
        return self.executor.map(_mp_invoke_worker, *iterables)

    def shutdown(self):
        self.executor.shutdown()
        os.environ.pop('_LK_IN_MP', 'yes')


class MPOpInvoker(ModelOpInvoker):
    def __init__(self, model, func, n_jobs, persist_method):
        if isinstance(model, PersistedModel):
            key = model
        else:
            key = persist(model, method=persist_method)
        func = pickle.dumps(func)
        ctx = LKContext.INSTANCE
        kid_tc = proc_count(level=1)
        os.environ['_LK_IN_MP'] = 'yes'
        _log.info('setting up multiprocessing.Pool w/ %d workers', n_jobs)
        self.pool = ctx.Pool(n_jobs, _initialize_mp_worker,
                             (key, func, kid_tc, log_queue(), get_root_seed()))

    def map(self, *iterables):
        return self.pool.starmap(_mp_invoke_worker, zip(*iterables))

    def shutdown(self):
        self.pool.close()
        os.environ.pop('_LK_IN_MP', 'yes')



import logging

from lenskit.sharing import persist, persist_binpickle
from lenskit.util.parallel import run_sp
from lenskit.util import Stopwatch

_log = logging.getLogger(__name__)


def _train_and_save(algo, file, ratings, kwargs):
    "Worker for subprocess model training"
    _log.info('training %s on %d ratings', algo, len(ratings))
    timer = Stopwatch()
    algo.fit(ratings, **kwargs)
    _log.info('trained %s in %s', algo, timer)
    if file is None:
        return persist_binpickle(algo).transfer()
    else:
        return persist_binpickle(algo, file=file).transfer()


def train_isolated(algo, ratings, *, file=None, **kwargs):
    """
    Train an algorithm in a subprocess to isolate the training process.  This
    function spawns a subprocess (in the same way that LensKit's multiprocessing
    support does), calls :meth:`lenskit.algorithms.Algorithm.fit` on it, and
    serializes the result for shared-memory use.

    Training the algorithm in a single-purpose subprocess makes sure that any
    training resources, such as TensorFlow sessions, are cleaned up by virtue
    of the process terminating when model training is completed.  It can also
    reduce memory use, because the original trained model and the shared memory
    version are not in memory at the same time.  While the batch functions use
    shared memory to reduce memory overhead for parallel processing, naive use
    of these functions will still have 2 copies of the model in memory, the
    shared one and the original, because the sharing process does not tear down
    the original model.  Training in a subprocess solves this problem elegantly.

    Args:
        algo(lenskit.algorithms.Algorithm):
            The algorithm to train.
        ratings(pandas.DataFrame):
            The rating data.
        file(str or pathlib.Path or None):
            The file in which to save the trained model.  If ``None``, uses a
            default file path or shared memory.
        kwargs(dict):
            Additional named parameters to :meth:`lenskit.algorithms.Algorithm.fit`.

    Returns:
        lenskit.sharing.PersistedObject:
            The saved model object.  This is the owner, so it needs to be closed
            when finished to free resources.
    """

    return run_sp(_train_and_save, algo, file, ratings, kwargs)



"""
TensorFlow-based algorithms.
"""

import logging

from .util import have_usable_tensorflow
from .biasedmf import BiasedMF      # noqa: F401
from .ibmf import IntegratedBiasMF  # noqa: F401
from .bpr import BPR                # noqa: F401

from lenskit.util.parallel import is_mp_worker

TF_AVAILABLE = have_usable_tensorflow()

_log = logging.getLogger(__name__)

if TF_AVAILABLE and is_mp_worker():
    _log.info('disabling GPUs in worker process')
    _tf.config.set_visible_devices([], 'GPU')



"""
Accumulator support.
"""

from numba import njit
from .array import swap


@njit
def _pair_downheap(pos: int, sp, limit, ks, vs):
    finished = False
    while not finished:
        min = pos
        left = 2*pos + 1
        right = 2*pos + 2
        if left < limit and vs[sp + left] < vs[sp + min]:
            min = left
        if right < limit and vs[sp + right] < vs[sp + min]:
            min = right
        if min != pos:
            # we want to swap!
            swap(vs, sp + pos, sp + min)
            swap(ks, sp + pos, sp + min)
            pos = min
        else:
            finished = True


@njit
def _pair_upheap(pos, sp, ks, vs):
    parent = (pos - 1) // 2
    while pos > 0 and vs[sp + parent] > vs[sp + pos]:
        swap(vs, sp + parent, sp + pos)
        swap(ks, sp + parent, sp + pos)
        pos = parent
        parent = (pos - 1) // 2


@njit
def kvp_minheap_insert(sp, ep, limit, k, v, keys, vals):
    """
    Insert a value (with key) into a heap-organized array subset, only keeping the top values.

    Args:
        sp(int): the start of the heap (inclusive)
        ep(int): the current end of the heap (exclusive)
        limit(int): the maximum size of the heap
        k: the key
        v: the value (used for sorting)
        keys: the key array, must be at least sp+limit.
        vals: the value array, same size as keys

    Returns:
        int: the new ep
    """

    if ep - sp < limit:
        # insert into heap without size problems
        # put on end, then upheap
        keys[ep] = k
        vals[ep] = v
        _pair_upheap(ep - sp, sp, keys, vals)
        return ep + 1

    elif v > vals[sp]:
        # heap is full, but new value is larger than old min
        # stick it on the front, and downheap
        keys[sp] = k
        vals[sp] = v
        _pair_downheap(0, sp, limit, keys, vals)
        return ep

    else:
        # heap is full and new value doesn't belong
        return ep


@njit
def kvp_minheap_sort(sp, ep, keys, vals):
    """
    Sort a heap-organized array subset by decreasing values.

    Args:
        sp(int): the start of the heap (inclusive).
        ep(int): the end of the heap (exclusive).
        keys: the key array
        vals: the value array
    """

    for i in range(ep-1, sp, -1):
        swap(keys, i, sp)
        swap(vals, i, sp)
        _pair_downheap(0, sp, i-sp, keys, vals)



"""
User-based k-NN collaborative filtering.
"""

from sys import intern
import logging

import pandas as pd
import numpy as np

from numba import njit

from .. import util, matrix
from . import Predictor
from ..util.accum import kvp_minheap_insert

_logger = logging.getLogger(__name__)


@njit
def _agg_weighted_avg(iur, item, sims, use):
    """
    Weighted-average aggregate.

    Args:
        iur(matrix._CSR): the item-user ratings matrix
        item(int): the item index in ``iur``
        sims(numpy.ndarray): the similarities for the users who have rated ``item``
        use(numpy.ndarray): positions in sims and the rating row to actually use
    """
    rates = iur.row_vs(item)
    num = 0.0
    den = 0.0
    for j in use:
        num += rates[j] * sims[j]
        den += np.abs(sims[j])
    return num / den


@njit
def _agg_sum(iur, item, sims, use):
    """
    Sum aggregate

    Args:
        iur(matrix._CSR): the item-user ratings matrix
        item(int): the item index in ``iur``
        sims(numpy.ndarray): the similarities for the users who have rated ``item``
        use(numpy.ndarray): positions in sims and the rating row to actually use
    """
    x = 0.0
    for j in use:
        x += sims[j]
    return x


@njit
def _score(items, results, iur, sims, nnbrs, min_sim, min_nbrs, agg):
    h_ks = np.empty(nnbrs, dtype=np.int32)
    h_vs = np.empty(nnbrs)
    used = np.zeros(len(results), dtype=np.int32)

    for i in range(len(results)):
        item = items[i]
        if item < 0:
            continue

        h_ep = 0

        # who has rated this item?
        i_users = iur.row_cs(item)

        # what are their similarities to our target user?
        i_sims = sims[i_users]

        # which of these neighbors do we really want to use?
        for j, s in enumerate(i_sims):
            if np.abs(s) < 1.0e-10:
                continue
            if min_sim is not None and s < min_sim:
                continue
            h_ep = kvp_minheap_insert(0, h_ep, nnbrs, j, s, h_ks, h_vs)

        if h_ep < min_nbrs:
            continue

        results[i] = agg(iur, item, i_sims, h_ks[:h_ep])
        used[i] = h_ep

    return used


class UserUser(Predictor):
    """
    User-user nearest-neighbor collaborative filtering with ratings. This user-user implementation
    is not terribly configurable; it hard-codes design decisions found to work well in the previous
    Java-based LensKit code.

    Args:
        nnbrs(int):
            the maximum number of neighbors for scoring each item (``None`` for unlimited)
        min_nbrs(int): the minimum number of neighbors for scoring each item
        min_sim(double): minimum similarity threshold for considering a neighbor
        center(bool):
            whether to normalize (mean-center) rating vectors.  Turn this off when working
            with unary data and other data types that don't respond well to centering.
        aggregate:
            the type of aggregation to do. Can be ``weighted-average`` or ``sum``.

    Attributes:
        user_index_(pandas.Index): User index.
        item_index_(pandas.Index): Item index.
        user_means_(numpy.ndarray): User mean ratings.
        rating_matrix_(matrix.CSR): Normalized user-item rating matrix.
        transpose_matrix_(matrix.CSR): Transposed un-normalized rating matrix.
    """
    AGG_SUM = intern('sum')
    AGG_WA = intern('weighted-average')

    def __init__(self, nnbrs, min_nbrs=1, min_sim=0, center=True, aggregate='weighted-average'):
        self.nnbrs = nnbrs
        self.min_nbrs = min_nbrs
        self.min_sim = min_sim
        self.center = center
        self.aggregate = intern(aggregate)

    def fit(self, ratings, **kwargs):
        """
        "Train" a user-user CF model.  This memorizes the rating data in a format that is usable
        for future computations.

        Args:
            ratings(pandas.DataFrame): (user, item, rating) data for collaborative filtering.

        Returns:
            UUModel: a memorized model for efficient user-based CF computation.
        """

        uir, users, items = matrix.sparse_ratings(ratings)

        # mean-center ratings
        if self.center:
            umeans = uir.normalize_rows('center')
        else:
            umeans = None

        # compute centered transpose
        iur = uir.transpose()

        # L2-normalize ratings so dot product is cosine
        if uir.values is None:
            uir.values = np.full(uir.nnz, 1.0)
        uir.normalize_rows('unit')

        mkl = matrix.mkl_ops()
        mkl_m = mkl.SparseM.from_csr(uir) if mkl else None

        self.rating_matrix_ = uir
        self.user_index_ = users
        self.user_means_ = umeans
        self.item_index_ = items
        self.transpose_matrix_ = iur
        self._mkl_m_ = mkl_m

        return self

    def predict_for_user(self, user, items, ratings=None):
        """
        Compute predictions for a user and items.

        Args:
            user: the user ID
            items (array-like): the items to predict
            ratings (pandas.Series):
                the user's ratings (indexed by item id); if provided, will be used to
                recompute the user's bias at prediction time.

        Returns:
            pandas.Series: scores for the items, indexed by item id.
        """

        watch = util.Stopwatch()
        items = pd.Index(items, name='item')

        ratings, umean = self._get_user_data(user, ratings)
        if ratings is None:
            return pd.Series(index=items, dtype='float64')
        assert len(ratings) == len(self.item_index_)  # ratings is a dense vector

        # now ratings is normalized to be a mean-centered unit vector
        # this means we can dot product to score neighbors
        # score the neighbors!
        if self._mkl_m_:
            nsims = np.zeros(len(self.user_index_))
            nsims = self._mkl_m_.mult_vec(1, ratings, 0, nsims)
        else:
            rmat = self.rating_matrix_.to_scipy()
            nsims = rmat @ ratings
        assert len(nsims) == len(self.user_index_)
        if user in self.user_index_:
            nsims[self.user_index_.get_loc(user)] = 0

        _logger.debug('computed user similarities')

        results = np.full(len(items), np.nan, dtype=np.float_)
        ri_pos = self.item_index_.get_indexer(items.values)
        if self.aggregate == self.AGG_WA:
            agg = _agg_weighted_avg
        elif self.aggregate == self.AGG_SUM:
            agg = _agg_sum
        else:
            raise ValueError('invalid aggregate ' + self.aggregate)

        _score(ri_pos, results, self.transpose_matrix_.N, nsims,
               self.nnbrs, self.min_sim, self.min_nbrs, agg)
        results += umean

        results = pd.Series(results, index=items, name='prediction')

        _logger.debug('scored %d of %d items for %s in %s',
                      results.notna().sum(), len(items), user, watch)
        return results

    def _get_user_data(self, user, ratings):
        "Get a user's data for user-user CF"
        rmat = self.rating_matrix_

        if ratings is None:
            try:
                upos = self.user_index_.get_loc(user)
                ratings = rmat.row(upos)
                umean = self.user_means_[upos] if self.user_means_ is not None else 0
            except KeyError:
                _logger.warning('user %d has no ratings and none provided', user)
                return None, 0
        else:
            _logger.debug('using provided ratings for user %d', user)
            if self.center:
                umean = ratings.mean()
                ratings = ratings - umean
            else:
                umean = 0
            unorm = np.linalg.norm(ratings)
            ratings = ratings / unorm
            ratings = ratings.reindex(self.item_index_, fill_value=0).values

        return ratings, umean

    def __getstate__(self):
        state = dict(self.__dict__)
        if '_mkl_m_' in state:
            del state['_mkl_m_']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.aggregate = intern(self.aggregate)
        mkl = matrix.mkl_ops()
        self._mkl_m_ = mkl.SparseM.from_csr(self.rating_matrix_) if mkl else None

    def __str__(self):
        return 'UserUser(nnbrs={}, min_sim={})'.format(self.nnbrs, self.min_sim)



"""
Support for sharing and saving models and data structures.
"""

import os
import warnings
from abc import abstractmethod, ABC
from contextlib import contextmanager
import threading
import logging

_log = logging.getLogger(__name__)

_store_state = threading.local()


def _save_mode():
    return getattr(_store_state, 'mode', 'save')


@contextmanager
def sharing_mode():
    """
    Context manager to tell models that pickling will be used for cross-process
    sharing, not model persistence.
    """
    old = _save_mode()
    _store_state.mode = 'share'
    try:
        yield
    finally:
        _store_state.mode = old


def in_share_context():
    """
    Query whether sharing mode is active.  If ``True``, we are currently in a
    :func:`sharing_mode` context, which means model pickling will be used for
    cross-process sharing.
    """
    return _save_mode() == 'share'


class PersistedModel(ABC):
    """
    A persisted model for inter-process model sharing.

    These objects can be pickled for transmission to a worker process.

    .. note::
        Subclasses need to override the pickling protocol to implement the
        proper pickling implementation.
    """

    @abstractmethod
    def get(self):
        """
        Get the persisted model, reconstructing it if necessary.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Release the persisted model resources.  Should only be called in the
        parent process (will do nothing in a child process).
        """
        pass

    def transfer(self):
        """
        Mark an object for ownership transfer.  This object, when pickled, will
        unpickle into an owning model that frees resources when closed. Used to
        transfer ownership of shared memory resources from child processes to
        parent processes.  Such an object should only be unpickled once.

        The default implementation sets the ``is_owner`` attribute to ``'transfer'``.

        Returns:
            ``self`` (for convenience)
        """
        if not self.is_owner:
            warnings.warning('non-owning objects should not be transferred', stacklevel=1)
        else:
            self.is_owner = 'transfer'
        return self


def persist(model, *, method=None):
    """
    Persist a model for cross-process sharing.

    This will return a persiste dmodel that can be used to reconstruct the model
    in a worker process (using :func:`reconstruct`).

    If no method is provided, this function automatically selects a model persistence
    strategy from the the following, in order:

    1. If `LK_TEMP_DIR` is set, use :mod:`binpickle` in shareable mode to save
       the object into the LensKit temporary directory.
    2. If :mod:`multiprocessing.shared_memory` is available, use :mod:`pickle`
       to save the model, placing the buffers into shared memory blocks.
    3. Otherwise, use :mod:`binpickle` in shareable mode to save the object
       into the system temporary directory.

    Args:
        model(obj):
            The model to persist.
        method(str or None):
            The method to use.  Can be one of ``binpickle`` or ``shm``.

    Returns:
        PersistedModel: The persisted object.
    """
    if method is not None:
        if method == 'binpickle':
            method = persist_binpickle
        elif method == 'shm':
            method = persist_shm
        elif not hasattr(method, '__call__'):
            raise ValueError('invalid method %s: must be one of binpickle, shm, or a funciton')

    if method is None:
        if SHM_AVAILABLE and 'LK_TEMP_DIR' not in os.environ:
            method = persist_shm
        else:
            method = persist_binpickle

    return method(model)


from .binpickle import persist_binpickle, BPKPersisted     # noqa: E402,F401
from .shm import persist_shm, SHMPersisted, SHM_AVAILABLE  # noqa: E402,F401



"""
Top-N evaluation metrics.
"""

import numpy as np


def precision(recs, truth):
    """
    Compute recommendation precision.
    """
    nrecs = len(recs)
    if nrecs == 0:
        return None

    ngood = recs['item'].isin(truth.index).sum()
    return ngood / nrecs


def recall(recs, truth):
    """
    Compute recommendation recall.
    """
    nrel = len(truth)
    if nrel == 0:
        return None

    ngood = recs['item'].isin(truth.index).sum()
    return ngood / nrel


def recip_rank(recs, truth):
    """
    Compute the reciprocal rank of the first relevant item in a list of recommendations.

    If no elements are relevant, the reciprocal rank is 0.
    """
    good = recs['item'].isin(truth.index)
    npz, = np.nonzero(good.to_numpy())
    if len(npz):
        return 1.0 / (npz[0] + 1.0)
    else:
        return 0.0


def _dcg(scores, discount=np.log2):
    """
    Compute the Discounted Cumulative Gain of a series of recommended items with rating scores.
    These should be relevance scores; they can be :math:`{0,1}` for binary relevance data.

    This is not a true top-N metric, but is a utility function for other metrics.

    Args:
        scores(array-like):
            The utility scores of a list of recommendations, in recommendation order.
        discount(ufunc):
            the rank discount function.  Each item's score will be divided the discount of its rank,
            if the discount is greater than 1.

    Returns:
        double: the DCG of the scored items.
    """
    scores = np.nan_to_num(scores)
    ranks = np.arange(1, len(scores) + 1)
    disc = discount(ranks)
    np.maximum(disc, 1, out=disc)
    np.reciprocal(disc, out=disc)
    return np.dot(scores, disc)


# @profile
def ndcg(recs, truth, discount=np.log2):
    """
    Compute the normalized discounted cumulative gain.

    Discounted cumultative gain is computed as:

    .. math::
        \\begin{align*}
        \\mathrm{DCG}(L,u) & = \\sum_{i=1}^{|L|} \\frac{r_{ui}}{d(i)}
        \\end{align*}

    This is then normalized as follows:

    .. math::
        \\begin{align*}
        \\mathrm{nDCG}(L, u) & = \\frac{\\mathrm{DCG}(L,u)}{\\mathrm{DCG}(L_{\\mathrm{ideal}}, u)}
        \\end{align*}

    Args:
        recs: The recommendation list.
        truth: The user's test data.
        discount(ufunc):
            The rank discount function.  Each item's score will be divided the discount of its rank,
            if the discount is greater than 1.
    """

    tpos = truth.index.get_indexer(recs['item'])
    tgood = tpos >= 0
    if 'rating' in truth.columns:
        i_rates = np.sort(truth.rating.values)[::-1]
        ideal = _dcg(i_rates, discount)
        # make an array of ratings for this rec list
        r_rates = truth['rating'].values[tpos]
        r_rates[tpos < 0] = 0
        achieved = _dcg(r_rates, discount)
    else:
        ideal = _dcg(np.ones(len(truth)), discount)
        achieved = _dcg(tgood, discount)

    return achieved / ideal



"""
Data set cross-folding.
"""

from collections import namedtuple
import logging
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from . import util

TTPair = namedtuple('TTPair', ['train', 'test'])
TTPair.__doc__ = 'Train-test pair (named tuple).'
TTPair.train.__doc__ = 'Train data for this pair.'
TTPair.test.__doc__ = 'Test data for this pair.'

_logger = logging.getLogger(__name__)


def partition_rows(data, partitions, *, rng_spec=None):
    """
    Partition a frame of ratings or other datainto train-test partitions.  This function does not
    care what kind of data is in `data`, so long as it is a Pandas DataFrame (or equivalent).

    Args:
        data(pandas.DataFrame):
            Ratings or other data you wish to partition.
        partitions(int):
            The number of partitions to produce.
        rng_spec:
            The random number generator or seed (see :py:func:`lenskit.util.rng`).

    Returns:
        iterator: an iterator of train-test pairs
    """
    _logger.info('partitioning %d ratings into %d partitions', len(data), partitions)

    # create an array of indexes
    rows = np.arange(len(data))
    # shuffle the indices & split into partitions
    rng = util.rng(rng_spec)
    rng.shuffle(rows)
    test_sets = np.array_split(rows, partitions)

    # convert each partition into a split
    for i, ts in enumerate(test_sets):
        test = data.iloc[ts, :]
        trains = test_sets[:i] + test_sets[(i + 1):]
        train_idx = np.concatenate(trains)
        train = data.iloc[train_idx, :]
        yield TTPair(train, test)


def sample_rows(data, partitions, size, disjoint=True, *, rng_spec=None):
    """
    Sample train-test a frame of ratings into train-test partitions.  This function does not care
    what kind of data is in `data`, so long as it is a Pandas DataFrame (or equivalent).

    We can loop over a sequence of train-test pairs::

        >>> from lenskit import datasets
        >>> ratings = datasets.MovieLens('data/ml-latest-small').ratings
        >>> for train, test in sample_rows(ratings, 5, 1000):
        ...     print(len(test))
        1000
        1000
        1000
        1000
        1000

    Sometimes for testing, it is useful to just get a single pair::

        >>> train, test = sample_rows(ratings, None, 1000)
        >>> len(test)
        1000
        >>> len(test) + len(train) - len(ratings)
        0

    Args:
        data(pandas.DataFrame):
            Data frame containing ratings or other data to partition.
        partitions(int or None):
            The number of partitions to produce.  If ``None``, produce a _single_ train-test
            pair instead of an iterator or list.
        size(int):
            The size of each sample.
        disjoint(bool):
            If ``True``, force samples to be disjoint.
        rng_spec:
            The random number generator or seed (see :py:func:`lenskit.util.rng`).

    Returns:
        iterator: An iterator of train-test pairs.
    """

    if partitions is None:
        test = data.sample(n=size)
        tr_mask = pd.Series(True, index=data.index)
        tr_mask.loc[test.index] = False
        train = data[tr_mask]
        return TTPair(train, test)

    if disjoint and partitions * size >= len(data):
        _logger.warning('wanted %d disjoint splits of %d each, but only have %d rows; partitioning',
                        partitions, size, len(data))
        return partition_rows(data, partitions)

    # create an array of indexes
    rows = np.arange(len(data))

    rng = util.rng(rng_spec)

    if disjoint:
        _logger.info('creating %d disjoint samples of size %d', partitions, size)
        ips = _disjoint_sample(rows, partitions, size, rng)

    else:
        _logger.info('taking %d samples of size %d', partitions, size)
        ips = _n_samples(rows, partitions, size, rng)

    return (TTPair(data.iloc[ip.train, :], data.iloc[ip.test, :]) for ip in ips)


def _disjoint_sample(xs, n, size, rng):
    # shuffle the indices & split into partitions
    rng.shuffle(xs)

    # convert each partition into a split
    for i in range(n):
        start = i * size
        test = xs[start:start + size]
        train = np.concatenate((xs[:start], xs[start + size:]))
        yield TTPair(train, test)


def _n_samples(xs, n, size, rng):
    for i in range(n):
        test = rng.choice(xs, size, False)
        train = np.setdiff1d(xs, test, assume_unique=True)
        yield TTPair(train, test)


class PartitionMethod(ABC):
    """
    Partition methods select test rows for a user or item.  Partition methods
    are callable; when called with a data frame, they return the test rows.
    """

    @abstractmethod
    def __call__(self, udf):
        """
        Subset a data frame.

        Args:
            udf(pandas.DataFrame):
                The input data frame of rows for a user or item.

        Returns:
            pandas.DataFrame:
                The data frame of test rows, a subset of ``udf``.
        """
        pass


class SampleN(PartitionMethod):
    """
    Randomly select a fixed number of test rows per user/item.

    Args:
        n(int): the number of test items to select
        rng: the random number generator or seed
    """

    def __init__(self, n, rng_spec=None):
        self.n = n
        self.rng = util.rng(rng_spec, legacy=True)

    def __call__(self, udf):
        return udf.sample(n=self.n, random_state=self.rng)


class SampleFrac(PartitionMethod):
    """
    Randomly select a fraction of test rows per user/item.

    Args:
        frac(float): the fraction items to select for testing.
    """
    def __init__(self, frac, rng_spec=None):
        self.fraction = frac
        self.rng = util.rng(rng_spec, legacy=True)

    def __call__(self, udf):
        return udf.sample(frac=self.fraction, random_state=self.rng)


class LastN(PartitionMethod):
    """
    Select a fixed number of test rows per user/item, based on ordering by a
    column.

    Args:
        n(int): The number of test items to select.
    """

    def __init__(self, n, col='timestamp'):
        self.n = n
        self.column = col

    def __call__(self, udf):
        return udf.sort_values(self.column).iloc[-self.n:]


class LastFrac(PartitionMethod):
    """
    Select a fraction of test rows per user/item.

    Args:
        frac(double): the fraction of items to select for testing.
    """
    def __init__(self, frac, col='timestamp'):
        self.fraction = frac
        self.column = col

    def __call__(self, udf):
        n = round(len(udf) * self.fraction)
        return udf.sort_values(self.column).iloc[-n:]


def partition_users(data, partitions: int, method: PartitionMethod, *, rng_spec=None):
    """
    Partition a frame of ratings or other data into train-test partitions user-by-user.
    This function does not care what kind of data is in `data`, so long as it is a Pandas DataFrame
    (or equivalent) and has a `user` column.

    Args:
        data(pandas.DataFrame): a data frame containing ratings or other data you wish to partition.
        partitions(int): the number of partitions to produce
        method(PartitionMethod): The method for selecting test rows for each user.
        rng_spec: The random number generator or seed (see :py:func:`lenskit.util.rng`).

    Returns
        iterator: an iterator of train-test pairs
    """

    user_col = data['user']
    users = user_col.unique()
    _logger.info('partitioning %d rows for %d users into %d partitions',
                 len(data), len(users), partitions)

    # create an array of indexes into user row
    rows = np.arange(len(users))
    # shuffle the indices & split into partitions
    rng = util.rng(rng_spec, legacy=True)
    rng.shuffle(rows)
    test_sets = np.array_split(rows, partitions)

    # convert each partition into a split
    for i, ts in enumerate(test_sets):
        # get our users!
        test_us = users[ts]
        # sample the data frame
        _logger.info('fold %d: selecting test ratings', i)
        ugf = data[data.user.isin(test_us)].groupby('user')
        test = ugf.apply(method)
        # get rid of the group index
        test = test.reset_index(0, drop=True)
        # now test is indexed on the data frame! so we can get the rest
        _logger.info('fold %d: partitioning training data', i)
        mask = pd.Series(True, index=data.index)
        mask[test.index] = False
        train = data[mask]
        yield TTPair(train, test)


def sample_users(data, partitions: int, size: int, method: PartitionMethod, disjoint=True, *,
                 rng_spec=None):
    """
    Create train-test partitions by sampling users.
    This function does not care what kind of data is in `data`, so long as it is
    a Pandas DataFrame (or equivalent) and has a `user` column.

    Args:
        data(pandas.DataFrame):
            Data frame containing ratings or other data you wish to partition.
        partitions(int):
            The number of partitions.
        size(int):
            The sample size.
        method(PartitionMethod):
            The method for obtaining user test ratings.
        rng_spec:
            The random number generator or seed (see :py:func:`lenskit.util.rng`).

    Returns:
        iterator: An iterator of train-test pairs (as :class:`TTPair` objects).
    """

    rng = util.rng(rng_spec, legacy=True)

    user_col = data['user']
    users = user_col.unique()
    if disjoint and partitions * size >= len(users):
        _logger.warning('cannot take %d disjoint samples of size %d from %d users',
                        partitions, size, len(users))
        yield from partition_users(data, partitions, method)
        return

    _logger.info('sampling %d users into %d partitions (n=%d)',
                 len(users), partitions, size)

    if disjoint:
        rng.shuffle(users)

    # generate our samples
    for i in range(partitions):
        # get our test users!
        if disjoint:
            test_us = users[i*size:(i+1)*size]
        else:
            test_us = rng.choice(users, size, False)

        # sample the data frame
        test = data[data.user.isin(test_us)].groupby('user').apply(method)
        # get rid of the group index
        test = test.reset_index(0, drop=True)
        # now test is indexed on the data frame! so we can get the rest
        rest = data.index.difference(test.index)
        train = data.loc[rest]
        yield TTPair(train, test)



"""
Basic utility algorithms and combiners.
"""

import logging
from collections.abc import Iterable, Sequence

import pandas as pd
import numpy as np

from ..matrix import sparse_ratings
from . import Predictor, Recommender, CandidateSelector
from ..util import derivable_rng

from .bias import Bias

_logger = logging.getLogger(__name__)


class Popular(Recommender):
    """
    Recommend the most popular items.

    Args:
        selector(CandidateSelector):
            The candidate selector to use. If ``None``, uses a new
            :class:`UnratedItemCandidateSelector`.

    Attributes:
        item_pop_(pandas.Series):
            Item rating counts (popularity)
    """

    def __init__(self, selector=None):
        if selector is None:
            self.selector = UnratedItemCandidateSelector()
        else:
            self.selector = selector

    def fit(self, ratings, **kwargs):
        pop = ratings.groupby('item').user.count()
        pop.name = 'score'
        self.item_pop_ = pop.astype('float64')
        self.selector.fit(ratings)

        return self

    def recommend(self, user, n=None, candidates=None, ratings=None):
        scores = self.item_pop_
        if candidates is None:
            candidates = self.selector.candidates(user, ratings)

        idx = scores.index.get_indexer(candidates)
        idx = idx[idx >= 0]
        scores = scores.iloc[idx]

        if n is None:
            return scores.sort_values(ascending=False).reset_index()
        else:
            return scores.nlargest(n).reset_index()

    def __str__(self):
        return 'Popular'


class Memorized(Predictor):
    """
    The memorized algorithm memorizes socres provided at construction time.
    """

    def __init__(self, scores):
        """
        Args:
            scores(pandas.DataFrame): the scores to memorize.
        """

        self.scores = scores

    def fit(self, *args, **kwargs):
        return self

    def predict_for_user(self, user, items, ratings=None):
        uscores = self.scores[self.scores.user == user]
        urates = uscores.set_index('item').rating
        return urates.reindex(items)


class Fallback(Predictor):
    """
    The Fallback algorithm predicts with its first component, uses the second to fill in
    missing values, and so forth.
    """

    def __init__(self, algorithms, *others):
        """
        Args:
            algorithms: a list of component algorithms.  Each one will be trained.
            others:
                additional algorithms, in which case ``algorithms`` is taken to be
                a single algorithm.
        """
        if others:
            self.algorithms = [algorithms] + list(others)
        elif isinstance(algorithms, Iterable) or isinstance(algorithms, Sequence):
            self.algorithms = algorithms
        else:
            self.algorithms = [algorithms]

    def fit(self, ratings, **kwargs):
        for algo in self.algorithms:
            algo.fit(ratings, **kwargs)

        return self

    def predict_for_user(self, user, items, ratings=None):
        remaining = pd.Index(items)
        preds = None

        for algo in self.algorithms:
            _logger.debug('predicting for %d items for user %s', len(remaining), user)
            aps = algo.predict_for_user(user, remaining, ratings=ratings)
            aps = aps[aps.notna()]
            if preds is None:
                preds = aps
            else:
                preds = pd.concat([preds, aps])
            remaining = remaining.difference(preds.index)
            if len(remaining) == 0:
                break

        return preds.reindex(items)

    def __str__(self):
        str_algos = [str(algo) for algo in self.algorithms]
        return 'Fallback([{}])'.format(', '.join(str_algos))


class TopN(Recommender, Predictor):
    """
    Basic recommender that implements top-N recommendation using a predictor.

    .. note::
        This class does not do anything of its own in :meth:`fit`.  If its
        predictor and candidate selector are both fit, the top-N recommender
        does not need to be fit.

    Args:
        predictor(Predictor):
            The underlying predictor.
        selector(CandidateSelector):
            The candidate selector.  If ``None``, uses :class:`UnratedItemCandidateSelector`.
    """

    def __init__(self, predictor, selector=None):
        self.predictor = predictor
        self.selector = selector if selector is not None else UnratedItemCandidateSelector()

    def fit(self, ratings, **kwargs):
        """
        Fit the recommender.

        Args:
            ratings(pandas.DataFrame):
                The rating or interaction data.  Passed changed to the predictor and
                candidate selector.
            args, kwargs:
                Additional arguments for the predictor to use in its training process.
        """
        self.predictor.fit(ratings, **kwargs)
        self.selector.fit(ratings, **kwargs)
        return self

    def fit_iters(self, ratings, **kwargs):
        if not hasattr(self.predictor, 'fit_iters'):
            raise AttributeError('predictor has no method fit_iters')

        self.selector.fit(ratings, **kwargs)
        pred = self.predictor
        for p in pred.fit_iters(ratings, **kwargs):
            self.predictor = p
            yield self

        self.predictor = pred

    def recommend(self, user, n=None, candidates=None, ratings=None):
        if candidates is None:
            candidates = self.selector.candidates(user, ratings)

        scores = self.predictor.predict_for_user(user, candidates, ratings)
        scores = scores[scores.notna()]
        if n is not None:
            scores = scores.nlargest(n)
        else:
            scores = scores.sort_values(ascending=False)
        scores.name = 'score'
        scores.index.name = 'item'
        return scores.reset_index()

    def predict(self, pairs, ratings=None):
        return self.predictor.predict(pairs, ratings)

    def predict_for_user(self, user, items, ratings=None):
        return self.predictor.predict_for_user(user, items, ratings)

    def __str__(self):
        return 'TopN/' + str(self.predictor)


class UnratedItemCandidateSelector(CandidateSelector):
    """
    :class:`CandidateSelector` that selects items a user has not rated as
    candidates.  When this selector is fit, it memorizes the rated items.

    Attributes:
        items_(pandas.Index): All known items.
        users_(pandas.Index): All known users.
        user_items_(CSR):
            Items rated by each known user, as positions in the ``items`` index.
    """
    items_ = None
    users_ = None
    user_items_ = None

    def fit(self, ratings, **kwargs):
        r2 = ratings[['user', 'item']]
        sparse = sparse_ratings(r2)
        _logger.info('trained unrated candidate selector for %d ratings', sparse.matrix.nnz)
        self.items_ = sparse.items
        self.users_ = sparse.users
        self.user_items_ = sparse.matrix

        return self

    def candidates(self, user, ratings=None):
        if ratings is None:
            try:
                uidx = self.users_.get_loc(user)
                uis = self.user_items_.row_cs(uidx)
            except KeyError:
                uis = None
        else:
            uis = self.items_.get_indexer(self.rated_items(ratings))
            uis = uis[uis >= 0]

        if uis is not None:
            mask = np.full(len(self.items_), True)
            mask[uis] = False
            return self.items_.values[mask]
        else:
            return self.items_.values


class Random(Recommender):
    """
    A random-item recommender.

    Attributes:
        selector(CandidateSelector):
            Selects candidate items for recommendation.
            Default is :class:`UnratedItemCandidateSelector`.
        rng_spec:
            Seed or random state for generating recommendations.  Pass
            ``'user'`` to deterministically derive per-user RNGS from
            the user IDs for reproducibility.
    """

    def __init__(self, selector=None, rng_spec=None):
        if selector:
            self.selector = selector
        else:
            self.selector = UnratedItemCandidateSelector()
        # Get a Pandas-compatible RNG
        self.rng_source = derivable_rng(rng_spec, legacy=True)
        self.items = None

    def fit(self, ratings, **kwargs):
        self.selector.fit(ratings, **kwargs)
        items = pd.DataFrame(ratings['item'].unique(), columns=['item'])
        self.items = items
        return self

    def recommend(self, user, n=None, candidates=None, ratings=None):
        if candidates is None:
            candidates = self.selector.candidates(user, ratings)
        if n is None:
            n = len(candidates)

        rng = self.rng_source(user)
        c_df = pd.DataFrame(candidates, columns=['item'])
        recs = c_df.sample(n, random_state=rng)
        return recs.reset_index(drop=True)

    def __str__(self):
        return 'Random'


class KnownRating(Predictor):
    """
    The known rating algorithm memorizes ratings provided in the fit method.
    """

    def fit(self, ratings, **kwargs):
        self.ratings = ratings.set_index(['user', 'item']).sort_index()
        return self

    def predict_for_user(self, user, items, ratings=None):
        uscores = self.ratings.xs(user, level='user', drop_level=True)
        return uscores.rating.reindex(items)



import logging

import pandas as pd

from .mf_common import MFPredictor

_logger = logging.getLogger(__name__)


class HPF(MFPredictor):
    """
    Hierarchical Poisson factorization, provided by
    `hpfrec <https://hpfrec.readthedocs.io/en/latest/>`_.

    Args:
        features(int): the number of features
        **kwargs: arguments passed to :py:class:`hpfrec.HPF`.
    """

    def __init__(self, features, **kwargs):
        self.features = features
        self._kwargs = kwargs

    def fit(self, ratings, **kwargs):
        import hpfrec

        users = pd.Index(ratings.user.unique())
        items = pd.Index(ratings.item.unique())

        if 'rating' in ratings.columns:
            count = ratings.rating.values.copy()
        else:
            _logger.info('no ratings found, assuming 1.0')
            count = 1.0

        hpfdf = pd.DataFrame({
            'UserId': users.get_indexer(ratings.user),
            'ItemId': items.get_indexer(ratings.item),
            'Count': count
        })

        hpf = hpfrec.HPF(self.features, reindex=False, **self._kwargs)

        _logger.info('fitting HPF model with %d features', self.features)
        hpf.fit(hpfdf)

        self.user_index_ = users
        self.item_index_ = items
        self.user_features_ = hpf.Theta
        self.item_features_ = hpf.Beta

        return self

    def predict_for_user(self, user, items, ratings=None):
        # look up user index
        return self.score_by_ids(user, items)



import logging

try:
    import tensorflow as tf
    import tensorflow.keras as k
except ImportError:
    tf = None

from lenskit import util
from ..mf_common import MFPredictor
from ..bias import Bias
from .util import init_tf_rng, check_tensorflow

_log = logging.getLogger(__name__)


class BiasedMF(MFPredictor):
    """
    Biased matrix factorization model for explicit feedback, optimized with
    TensorFlow.

    This is a basic TensorFlow implementation of the biased matrix factorization
    model for rating prediction:

    .. math::
        s(i|u) = b + b_u + b_i + \\vec{p}_u \\cdot \\vec{q_i}

    User and item embedding matrices are regularized with :math:`L_2` regularization,
    governed by a regularization term :math:`\\lambda`.  Regularizations for the user
    and item embeddings are then computed as follows:

    .. math::
        \\lambda_u = \\lambda / |U| \\\\
        \\lambda_i = \\lambda / |I| \\\\

    This rescaling allows the regularization term to be independent of the number of
    users and items.

    Because the model is very simple, this algorithm works best with large
    batch sizes.

    This implementation uses :class:`lenskit.algorithms.bias.Bias` for computing
    the biases, and uses TensorFlow to fit a matrix factorization on the residuals.
    It then extracts the resulting matrices, and relies on :class:`MFPredictor`
    to implement the prediction logic, like :class:`lenskit.algorithms.als.BiasedMF`.
    Its code is suitable as an example of how to build a Keras/TensorFlow algorithm
    implementation for LensKit where TF is only used in the train stage.

    A variety of resources informed the design, most notably `this one`_.

    .. _this one: https://towardsdatascience.com/1fba34180699

    Args:
        features(int): The number of latent features to learn.
        bias: The bias model to use.
        damping: The bias damping, if ``bias`` is ``True``.
        epochs(int): The number of epochs to train.
        batch_size(int): The Keras batch size.
        reg(double):
            The regularization term :math:`\\lambda` used to derive embedding vector
            regularizations.
        rng_spec:
            The random number generator initialization.
    """

    def __init__(self, features=50, *, bias=True, damping=5,
                 epochs=5, batch_size=10000, reg=0.02, rng_spec=None):
        check_tensorflow()
        self.features = features
        self.epochs = epochs
        self.batch_size = batch_size
        self.reg = reg
        if bias is True:
            self.bias = Bias(damping)
        else:
            self.bias = bias
        self.rng_spec = rng_spec

    def fit(self, ratings, **kwargs):
        timer = util.Stopwatch()
        normed = self.bias.fit_transform(ratings, indexes=True)
        model = self._build_model(len(self.bias.user_offsets_),
                                  len(self.bias.item_offsets_))

        _log.info('[%s] training model', timer)
        model.fit([normed['uidx'], normed['iidx']], normed['rating'],
                  epochs=self.epochs, batch_size=self.batch_size)

        _log.info('[%s] model finished, extracting weights', timer)
        self.user_features_ = model.get_layer('user-embed').get_weights()[0]
        self.item_features_ = model.get_layer('item-embed').get_weights()[0]

        self.user_index_ = self.bias.user_index
        self.item_index_ = self.bias.item_index

        return self

    def _build_model(self, n_users, n_items):
        n_features = self.features
        _log.info('configuring TensorFlow model for %d features from %d users and %d items',
                  n_features, n_users, n_items)

        init_tf_rng(self.rng_spec)

        # User input layer
        u_input = k.Input(shape=(1,), dtype='int32', name='user')
        # User embedding layer.
        u_reg = k.regularizers.l2(self.reg / n_users)
        u_embed = k.layers.Embedding(input_dim=n_users, output_dim=n_features, input_length=1,
                                     embeddings_regularizer=u_reg,
                                     embeddings_initializer='random_normal',
                                     name='user-embed')(u_input)
        # The embedding layer produces an extra dimension. Remove it.
        u_flat = k.layers.Flatten(name='user-vector')(u_embed)

        # Do the same thing for items
        i_input = k.Input(shape=(1,), dtype='int32', name='item')
        i_reg = k.regularizers.l2(self.reg / n_items)
        i_embed = k.layers.Embedding(input_dim=n_items, output_dim=n_features, input_length=1,
                                     embeddings_regularizer=i_reg,
                                     embeddings_initializer='random_normal',
                                     name='item-embed')(i_input)
        i_flat = k.layers.Flatten(name='item-vector')(i_embed)

        # Predict ratings using a dot product of users and items
        prod = k.layers.Dot(name='score', axes=1)([u_flat, i_flat])

        # Assemble the model and configure to optimize
        model = k.Model([u_input, i_input], prod, name='classic-mf')
        model.compile('adam', 'mean_squared_error')

        return model

    def predict_for_user(self, user, items, ratings=None):
        # look up user index
        preds = self.score_by_ids(user, items)
        preds = self.bias.inverse_transform_user(user, preds)
        return preds



import logging

import pandas as pd
import numpy as np
try:
    import tensorflow as tf
    import tensorflow.keras as k
except ImportError:
    tf = None

from lenskit import util
from .. import Predictor
from .util import init_tf_rng, check_tensorflow

_log = logging.getLogger(__name__)


if tf is not None:
    class ScoreLayer(k.layers.Layer):
        """
        Custom layer for scoring.  We use this so that we can have the global mean value
        as an isolated variable for TensorFlow to optimize. Optimization is more effective
        if we initialize this variable to the global average rating.
        """
        def __init__(self, mean, **kwargs):
            super().__init__(**kwargs)
            self._mean = mean
            self.global_bias = tf.Variable(mean)

        def call(self, ub, ib, dot):
            return self.global_bias + ub + ib + dot

        def get_config(self):
            return {'mean': self._mean}


class IntegratedBiasMF(Predictor):
    """
    Biased matrix factorization model for explicit feedback, optimizing both bias
    and embeddings with TensorFlow.

    This is a basic TensorFlow implementation of the biased matrix factorization
    model for rating prediction:

    .. math::
        s(i|u) = b + b_u + b_i + \\vec{p}_u \\cdot \\vec{q_i}

    User and item embedding matrices are regularized with :math:`L_2` regularization,
    governed by a regularization term :math:`\\lambda`.  Regularizations for the user
    and item embeddings are then computed as follows:

    .. math::
        \\lambda_u = \\lambda / |U| \\\\
        \\lambda_i = \\lambda / |I| \\\\

    This rescaling allows the regularization term to be independent of the number of
    users and items.  The same rescaling applies to the bias regularization.

    Because the model is very simple, this algorithm works best with large
    batch sizes.

    This implementation uses TensorFlow to fit the entire model, including user/item
    biases and residuals, and uses TensorFlow to do the final predictions as well.
    Its code is suitable as an example of how to build a Keras/TensorFlow algorithm
    implementation for LensKit where TF used for the entire process.

    A variety of resources informed the design, most notably `this one`_ and
    `Chin-chi Hsu's example code`_.

    .. _this one: https://towardsdatascience.com/1fba34180699
    .. _Chin-chi Hsu's code: https://github.com/chinchi-hsu/KerasCollaborativeFiltering

    Args:
        features(int): The number of latent features to learn.
        epochs(int): The number of epochs to train.
        batch_size(int): The Keras batch size.
        reg(double):
            The regularization term for the embedding vectors.
        bias_reg(double):
            The regularization term for the bias vectors.
        rng_spec:
            The random number generator initialization.

    Attributes:
        model: The Keras model.
    """

    model = None

    def __init__(self, features=50, *, epochs=5, batch_size=10000,
                 reg=0.02, bias_reg=0.2, rng_spec=None):
        check_tensorflow()
        self.features = features
        self.epochs = epochs
        self.batch_size = batch_size
        self.reg = reg
        self.bias_reg = bias_reg
        self.rng_spec = rng_spec

    def fit(self, ratings, **kwargs):
        timer = util.Stopwatch()

        users = pd.Index(np.unique(ratings['user']))
        items = pd.Index(np.unique(ratings['item']))

        u_no = users.get_indexer(ratings['user'])
        i_no = items.get_indexer(ratings['item'])
        mean = np.mean(ratings['rating'].values, dtype='f4')  # TensorFlow is using 32-bits

        model = self._build_model(len(users), len(items), mean)

        _log.info('[%s] training model', timer)
        model.fit([u_no, i_no], ratings['rating'],
                  epochs=self.epochs, batch_size=self.batch_size)

        _log.info('[%s] model finished', timer)

        self.user_index_ = users
        self.item_index_ = items
        self.model = model

        return self

    def _build_model(self, n_users, n_items, mean):
        n_features = self.features
        _log.info('configuring TensorFlow model for %d features from %d users and %d items',
                  n_features, n_users, n_items)
        _log.info('global mean rating is %f', mean)

        init_tf_rng(self.rng_spec)

        # User input layer
        u_input = k.Input(shape=(1,), dtype='int32', name='user')
        # User embedding layer.
        u_reg = k.regularizers.l2(self.reg / n_users)
        u_embed = k.layers.Embedding(input_dim=n_users, output_dim=n_features, input_length=1,
                                     embeddings_regularizer=u_reg,
                                     embeddings_initializer='random_normal',
                                     name='user-embed')(u_input)
        # The embedding layer produces an extra dimension. Remove it.
        u_flat = k.layers.Flatten(name='user-vector')(u_embed)
        # User bias layer - it's a 1-dimensional embedding.
        ub_reg = k.regularizers.l2(self.bias_reg / n_users)
        u_bias = k.layers.Embedding(input_dim=n_users, output_dim=1, input_length=1,
                                    embeddings_regularizer=ub_reg,
                                    embeddings_initializer='random_normal',
                                    name='user-bias')(u_input)
        ub_flat = k.layers.Flatten(name='user-bv')(u_bias)

        # Do the same thing for items
        i_input = k.Input(shape=(1,), dtype='int32', name='item')
        i_reg = k.regularizers.l2(self.reg / n_items)
        i_embed = k.layers.Embedding(input_dim=n_items, output_dim=n_features, input_length=1,
                                     embeddings_regularizer=i_reg,
                                     embeddings_initializer='random_normal',
                                     name='item-embed')(i_input)
        i_flat = k.layers.Flatten(name='item-vector')(i_embed)
        ib_reg = k.regularizers.l2(self.bias_reg / n_items)
        i_bias = k.layers.Embedding(input_dim=n_items, output_dim=1, input_length=1,
                                    embeddings_regularizer=ib_reg,
                                    embeddings_initializer='random_normal',
                                    name='item-bias')(i_input)
        ib_flat = k.layers.Flatten(name='item-bv')(i_bias)

        # Predict ratings using a dot product of users and items
        dot = k.layers.Dot(name='resid-score', axes=1)([u_flat, i_flat])
        # score = k.layers.Add(name='score')([u_bias, i_bias, dot])
        score = ScoreLayer(mean, name='score')(ub_flat, ib_flat, dot)

        # Assemble the model and configure to optimize
        model = k.Model([u_input, i_input], score, name='int-mf')
        model.compile('adam', 'mean_squared_error')

        return model

    def predict_for_user(self, user, items, ratings=None):
        if user not in self.user_index_:
            return pd.Series(np.nan, index=items)

        items = np.array(items)
        uno = self.user_index_.get_loc(user)
        inos = self.item_index_.get_indexer_for(items).astype('i4')
        good_inos = inos[inos >= 0]
        good_items = items[inos >= 0]
        unos = np.full(len(good_inos), uno, dtype='i4')
        _log.debug('predicting %d items for user %d', len(good_inos), user)

        ys = self.model.predict([unos, good_inos])

        res = pd.Series(ys[:, 0], index=good_items)
        return res.reindex(items)

    def __getstate__(self):
        state = dict(self.__dict__)
        if self.model is not None:
            # we need to save the model
            del state['model']
            _log.info('extracting model config and weights')
            state['model_config'] = self.model.get_config()
            state['model_weights'] = self.model.get_weights()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if 'model_config' in state:
            _log.info('rehydrating model')
            self.model = k.Model.from_config(state['model_config'], custom_objects={
                'ScoreLayer': ScoreLayer
            })
            self.model.set_weights(state['model_weights'])



import logging
import warnings

import pandas as pd
import numpy as np

from ..algorithms import Recommender
from .. import util
from ..sharing import PersistedModel

_logger = logging.getLogger(__name__)


def _recommend_user(algo, req):
    user, n, candidates = req

    _logger.debug('generating recommendations for %s', user)
    watch = util.Stopwatch()
    res = algo.recommend(user, n, candidates)
    _logger.debug('%s recommended %d/%s items for %s in %s',
                  str(algo), len(res), n, user, watch)

    res['user'] = user
    res['rank'] = np.arange(1, len(res) + 1)

    return res.reset_index(drop=True)


def __standard_cand_fun(candidates):
    """
    Convert candidates from the forms accepted by :py:fun:`recommend` into
    a standard form, a function that takes a user and returns a candidate
    list.
    """
    if isinstance(candidates, dict):
        return candidates.get
    elif candidates is None:
        return lambda u: None
    else:
        return candidates


def recommend(algo, users, n, candidates=None, *, n_jobs=None, **kwargs):
    """
    Batch-recommend for multiple users.  The provided algorithm should be a
    :py:class:`algorithms.Recommender`.

    Args:
        algo: the algorithm
        users(array-like): the users to recommend for
        n(int): the number of recommendations to generate (None for unlimited)
        candidates:
            the users' candidate sets. This can be a function, in which case it will
            be passed each user ID; it can also be a dictionary, in which case user
            IDs will be looked up in it.  Pass ``None`` to use the recommender's
            built-in candidate selector (usually recommended).
        n_jobs(int):
            The number of processes to use for parallel recommendations.  Passed to
            :func:`lenskit.util.parallel.invoker`.

    Returns:
        A frame with at least the columns ``user``, ``rank``, and ``item``; possibly also
        ``score``, and any other columns returned by the recommender.
    """

    if n_jobs is None and 'nprocs' in kwargs:
        n_jobs = kwargs['nprocs']
        warnings.warn('nprocs is deprecated, use n_jobs', DeprecationWarning)

    if not isinstance(algo, PersistedModel):
        rec_algo = Recommender.adapt(algo)
        if candidates is None and rec_algo is not algo:
            warnings.warn('no candidates provided and algo is not a recommender, unlikely to work')
        algo = rec_algo
        del rec_algo

    if 'ratings' in kwargs:
        warnings.warn('Providing ratings to recommend is not supported', DeprecationWarning)

    candidates = __standard_cand_fun(candidates)

    with util.parallel.invoker(algo, _recommend_user, n_jobs=n_jobs) as worker:
        _logger.info('recommending with %s for %d users (n_jobs=%s)',
                     str(algo), len(users), n_jobs)
        del algo
        timer = util.Stopwatch()
        results = worker.map((user, n, candidates(user)) for user in users)
        results = pd.concat(results, ignore_index=True, copy=False)
        _logger.info('recommended for %d users in %s', len(users), timer)

    return results
